# sage_setup: distribution = sagemath-categories
r"""
Jacobi elliptic functions

This module implements the 12 Jacobi elliptic functions, along with their
inverses and the Jacobi amplitude function.

Jacobi elliptic functions can be thought of as generalizations
of both ordinary and hyperbolic trig functions. There are twelve
Jacobian elliptic functions. Each of the twelve corresponds to an
arrow drawn from one corner of a rectangle to another.

::

                n ------------------- d
                |                     |
                |                     |
                |                     |
                s ------------------- c

Each of the corners of the rectangle are labeled, by convention, ``s``,
``c``, ``d``, and ``n``. The rectangle is understood to be lying on the complex
plane, so that ``s`` is at the origin, ``c`` is on the real axis, and ``n`` is
on the imaginary axis. The twelve Jacobian elliptic functions are
then `\operatorname{pq}(x)`, where ``p`` and ``q`` are one of the letters
``s``, ``c``, ``d``, ``n``.

The Jacobian elliptic functions are then the unique
doubly-periodic, meromorphic functions satisfying the following
three properties:

#. There is a simple zero at the corner ``p``, and a simple pole at the
   corner ``q``.
#. The step from ``p`` to ``q`` is equal to half the period of the function
   `\operatorname{pq}(x)`; that is, the function `\operatorname{pq}(x)` is
   periodic in the direction ``pq``, with the period being twice the distance
   from ``p`` to ``q``. `\operatorname{pq}(x)` is periodic in the other two
   directions as well, with a period such that the distance from ``p`` to one
   of the other corners is a quarter period.
#. If the function `\operatorname{pq}(x)` is expanded in terms of `x` at one of
   the corners, the leading term in the expansion has a coefficient of 1.
   In other words, the leading term of the expansion of `\operatorname{pq}(x)`
   at the corner ``p`` is `x`; the leading term of the expansion at the corner
   ``q`` is `1/x`, and the leading term of an expansion at the other two
   corners is 1.

We can write

.. MATH::

    \operatorname{pq}(x) = \frac{\operatorname{pr}(x)}{\operatorname{qr}(x)}

where ``p``, ``q``, and ``r`` are any of the
letters ``s``, ``c``, ``d``, ``n``, with
the understanding that `\mathrm{ss} = \mathrm{cc} = \mathrm{dd}
= \mathrm{nn} = 1`.

Let

.. MATH::

    u = \int_0^{\phi} \frac{d\theta} {\sqrt {1-m \sin^2 \theta}},

then the *Jacobi elliptic function* `\operatorname{sn}(u)` is given by

.. MATH::

    \operatorname{sn}{u} = \sin{\phi}

and `\operatorname{cn}(u)` is given by

.. MATH::

    \operatorname{cn}{u} = \cos{\phi}

and

.. MATH::

    \operatorname{dn}{u} = \sqrt{1 - m\sin^2 \phi}.

To emphasize the dependence on `m`, one can write
`\operatorname{sn}(u|m)` for example (and similarly for `\mathrm{cn}` and
`\mathrm{dn}`). This is the notation used below.

For a given `k` with `0 < k < 1` they therefore are
solutions to the following nonlinear ordinary differential
equations:

- `\operatorname{sn}\,(x;k)` solves the differential equations

  .. MATH::

      \frac{d^2 y}{dx^2} + (1+k^2) y - 2 k^2 y^3 = 0
      \quad \text{ and } \quad
      \left(\frac{dy}{dx}\right)^2 = (1-y^2) (1-k^2 y^2).

- `\operatorname{cn}(x;k)` solves the differential equations

  .. MATH::

      \frac{d^2 y}{dx^2} + (1-2k^2) y + 2 k^2 y^3 = 0
      \quad \text{ and } \quad
      \left(\frac{dy}{dx}\right)^2 = (1-y^2)(1-k^2 + k^2 y^2).

- `\operatorname{dn}(x;k)` solves the differential equations

  .. MATH::

      \frac{d^2 y}{dx^2} - (2 - k^2) y + 2 y^3 = 0
      \quad \text{ and } \quad
      \left(\frac{dy}{dx}\right)^2 = y^2 (1 - k^2 - y^2).

  If `K(m)` denotes the complete elliptic integral of the
  first kind (named ``elliptic_kc`` in Sage), the elliptic functions
  `\operatorname{sn}(x|m)` and `\operatorname{cn}(x|m)` have real periods
  `4K(m)`, whereas `\operatorname{dn}(x|m)` has a period
  `2K(m)`. The limit `m \rightarrow 0` gives
  `K(0) = \pi/2` and trigonometric functions:
  `\operatorname{sn}(x|0) = \sin{x}`, `\operatorname{cn}(x|0) = \cos{x}`,
  `\operatorname{dn}(x|0) = 1`. The limit `m \rightarrow 1` gives
  `K(1) \rightarrow \infty` and hyperbolic functions:
  `\operatorname{sn}(x|1) = \tanh{x}`,
  `\operatorname{cn}(x|1) = \operatorname{sech}{x}`,
  `\operatorname{dn}(x|1) = \operatorname{sech}{x}`.

REFERENCES:

- :wikipedia:`Jacobi%27s_elliptic_functions`

- [KS2002]_

AUTHORS:

- David Joyner (2006): initial version

- Eviatar Bach (2013): complete rewrite, new numerical evaluation, and
  addition of the Jacobi amplitude function
"""
# ****************************************************************************
#       Copyright (C) 2006 David Joyner <wdj@usna.edu>
#       Copyright (C) 2013 Eviatar Bach <eviatarbach@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#  as published by the Free Software Foundation; either version 2 of
#  the License, or (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************
from sage.functions.hyperbolic import (arctanh, arccosh, arcsinh, arcsech,
                                       arccsch, arccoth, cosh, coth, sech,
                                       csch, tanh, sinh)
from sage.functions.special import elliptic_e, elliptic_kc
from sage.functions.trig import (arctan, arcsin, arccos, arccot, arcsec,
                                 arccsc, csc, sec, sin, cos, tan, cot)
from sage.misc.lazy_import import lazy_import
from sage.rings.integer import Integer
from sage.rings.rational_field import QQ
from sage.symbolic.function import BuiltinFunction

lazy_import('sage.misc.latex', 'latex')

lazy_import('sage.libs.mpmath.utils', 'call', as_='_mpmath_utils_call')
lazy_import('mpmath', 'ellipfun', as_='_mpmath_ellipfun')

HALF = QQ((1, 2))


class Jacobi(BuiltinFunction):
    """
    Base class for the Jacobi elliptic functions.
    """
    def __init__(self, kind):
        r"""
        Initialize ``self``.

        EXAMPLES::

            sage: from sage.functions.jacobi import Jacobi
            sage: Jacobi('sn')
            jacobi_sn

        TESTS::

            sage: N(jacobi("sn", I, 1/2))   # abs tol 1e-12                             # needs sage.symbolic
            -8.59454886300046e-73 + 1.34737147138542*I

            sage: # optional - fricas, needs sage.symbolic
            sage: CN = fricas(jacobi('cn',x, 2)); CN
            jacobiCn(x,2)
            sage: fricas.series(CN, x=0)
                1  2   3  4   17  6    79  8    1381  10      11
            1 - - x  + - x  - -- x  + --- x  - ----- x   + O(x  )
                2      8      80      640      19200
            sage: fricas(jacobi('sn',x, 2))
            jacobiSn(x,2)
            sage: fricas(jacobi('dn',x, 2))
            jacobiDn(x,2)
        """
        if kind not in ['nd', 'ns', 'nc', 'dn', 'ds', 'dc', 'sn', 'sd',
                        'sc', 'cn', 'cd', 'cs']:
            raise ValueError("kind must be one of 'nd', 'ns', 'nc', 'dn', "
                             "'ds', 'dc', 'sn', 'sd', 'sc', 'cn', 'cd', 'cs'.")
        self.kind = kind
        BuiltinFunction.__init__(self,
            name=f'jacobi_{kind}', nargs=2, evalf_params_first=False,
            conversions=dict(maple=('Jacobi{}'.format(kind.upper())),
                             mathematica=('Jacobi{}'.format(kind.upper())),
                             fricas=('jacobi{}'.format(kind.capitalize())),
                             maxima=('jacobi_{}'.format(kind))))

    def _eval_(self, x, m):
        r"""
        TESTS:

        Check that the simplifications are correct::

            sage: # needs mpmath sage.symbolic
            sage: from sage.libs.mpmath.all import almosteq
            sage: almosteq(n(jacobi_nd(8, 0, hold=True)), n(jacobi_nd(8, 0)))
            True
            sage: almosteq(n(jacobi_nd(1, 1, hold=True)), n(jacobi_nd(1, 1)))
            True
            sage: almosteq(n(jacobi_nd(0, -5, hold=True)), n(jacobi_nd(0, -5)))
            True
            sage: almosteq(n(jacobi_ns(-4, 0, hold=True)), n(jacobi_ns(-4, 0)))
            True
            sage: almosteq(n(jacobi_ns(-2, 1, hold=True)), n(jacobi_ns(-2, 1)))
            True
            sage: almosteq(n(jacobi_nc(2, 0, hold=True)), n(jacobi_nc(2, 0)))
            True
            sage: almosteq(n(jacobi_nc(1, 1, hold=True)), n(jacobi_nc(1, 1)))
            True
            sage: almosteq(n(jacobi_nc(0, 0, hold=True)), n(jacobi_nc(0, 0)))
            True
            sage: almosteq(n(jacobi_dn(-10, 0, hold=True)), n(jacobi_dn(-10, 0)))
            True
            sage: almosteq(n(jacobi_dn(-1, 1, hold=True)), n(jacobi_dn(-1, 1)))
            True
            sage: almosteq(n(jacobi_dn(0, 3, hold=True)), n(jacobi_dn(0, 3)))
            True
            sage: almosteq(n(jacobi_ds(2, 0, hold=True)), n(jacobi_ds(2, 0)))
            True
            sage: almosteq(n(jacobi_dc(-1, 0, hold=True)), n(jacobi_dc(-1, 0)))
            True
            sage: almosteq(n(jacobi_dc(-8, 1, hold=True)), n(jacobi_dc(-8, 1)))
            True
            sage: almosteq(n(jacobi_dc(0, -10, hold=True)), n(jacobi_dc(0, -10)))
            True
            sage: almosteq(n(jacobi_sn(-7, 0, hold=True)), n(jacobi_sn(-7, 0)))
            True
            sage: almosteq(n(jacobi_sn(-3, 1, hold=True)), n(jacobi_sn(-3, 1)))
            True
            sage: almosteq(n(jacobi_sn(0, -6, hold=True)), n(jacobi_sn(0, -6)))
            True
            sage: almosteq(n(jacobi_sd(4, 0, hold=True)), n(jacobi_sd(4, 0)))
            True
            sage: almosteq(n(jacobi_sd(0, 1, hold=True)), n(jacobi_sd(0, 1)))
            True
            sage: almosteq(n(jacobi_sd(0, 3, hold=True)), n(jacobi_sd(0, 3)))
            True
            sage: almosteq(n(jacobi_sc(-9, 0, hold=True)), n(jacobi_sc(-9, 0)))
            True
            sage: almosteq(n(jacobi_sc(0, 1, hold=True)), n(jacobi_sc(0, 1)))
            True
            sage: almosteq(n(jacobi_sc(0, -10, hold=True)), n(jacobi_sc(0, -10)))
            True
            sage: almosteq(n(jacobi_cn(-2, 0, hold=True)), n(jacobi_cn(-2, 0)))
            True
            sage: almosteq(n(jacobi_cn(6, 1, hold=True)), n(jacobi_cn(6, 1)))
            True
            sage: almosteq(n(jacobi_cn(0, -10, hold=True)), n(jacobi_cn(0, -10)))
            True
            sage: almosteq(n(jacobi_cd(9, 0, hold=True)), n(jacobi_cd(9, 0)))
            True
            sage: almosteq(n(jacobi_cd(-8, 1, hold=True)), n(jacobi_cd(-8, 1)))
            True
            sage: almosteq(n(jacobi_cd(0, 1, hold=True)), n(jacobi_cd(0, 1)))
            True
            sage: almosteq(n(jacobi_cs(-9, 0, hold=True)), n(jacobi_cs(-9, 0)))
            True
            sage: almosteq(n(jacobi_cs(-6, 1, hold=True)), n(jacobi_cs(-6, 1)))
            True
        """
        if self.kind == 'nd':
            if m == 0:
                return Integer(1)
            elif m == 1:
                return cosh(x)
            elif x == 0:
                return Integer(1)
        elif self.kind == 'ns':
            if m == 0:
                return csc(x)
            elif m == 1:
                return coth(x)
        elif self.kind == 'nc':
            if m == 0:
                return sec(x)
            elif m == 1:
                return cosh(x)
            elif x == 0:
                return Integer(1)
        elif self.kind == 'dn':
            if m == 0:
                return Integer(1)
            elif m == 1:
                return sech(x)
            elif x == 0:
                return Integer(1)
        elif self.kind == 'ds':
            if m == 0:
                return csc(x)
        elif self.kind == 'dc':
            if m == 0:
                return sec(x)
            elif m == 1:
                return Integer(1)
            elif x == 0:
                return Integer(1)
        elif self.kind == 'sn':
            if m == 0:
                return sin(x)
            elif m == 1:
                return tanh(x)
            elif x == 0:
                return Integer(0)
        elif self.kind == 'sd':
            if m == 0:
                return sin(x)
            elif m == 1:
                return sinh(x)
            elif x == 0:
                return Integer(0)
        elif self.kind == 'sc':
            if m == 0:
                return tan(x)
            elif m == 1:
                return sinh(x)
            elif x == 0:
                return Integer(0)
        elif self.kind == 'cn':
            if m == 0:
                return cos(x)
            elif m == 1:
                return sech(x)
            elif x == 0:
                return Integer(1)
        elif self.kind == 'cd':
            if m == 0:
                return cos(x)
            elif m == 1:
                return Integer(1)
            elif x == 0:
                return Integer(1)
        elif self.kind == 'cs':
            if m == 0:
                return cot(x)
            elif m == 1:
                return csch(x)
        return

    def _evalf_(self, x, m, parent=None, algorithm=None):
        r"""
        TESTS::

            sage: jacobi_sn(3, 4).n(100)                                                # needs mpmath sage.symbolic
            -0.33260000892770027112809652714 + 1.7077912301715219199143891076e-33*I
            sage: jacobi_dn(I, I).n()                                                   # needs mpmath sage.symbolic
            0.874189950651018 + 0.667346865048825*I
        """
        return _mpmath_utils_call(_mpmath_ellipfun, self.kind, x, m, parent=parent)

    def _derivative_(self, x, m, diff_param):
        r"""
        TESTS:

        sn, cn, and dn are analytic for all real ``x``, so we can check
        that the derivatives are correct by computing the series::

            sage: # needs mpmath sage.symbolic
            sage: from mpmath import almosteq
            sage: a = 0.9327542442482303
            sage: b = 0.7402326293643771
            sage: almosteq(jacobi_sn(x, b).series(x, 10).subs(x=a),
            ....:          jacobi_sn(a, b), abs_eps=0.01)
            True
            sage: almosteq(jacobi_cn(x, b).series(x, 10).subs(x=a),
            ....:          jacobi_cn(a, b), abs_eps=0.01)
            True
            sage: almosteq(jacobi_dn(x, b).series(x, 10).subs(x=a),
            ....:          jacobi_dn(a, b), abs_eps=0.01)
            True
        """
        if diff_param == 0:
            # From Wolfram Functions Site
            if self.kind == 'cd':
                return (m - Integer(1)) * jacobi_nd(x, m) * jacobi_sd(x, m)
            elif self.kind == 'cn':
                return -jacobi_sn(x, m) * jacobi_dn(x, m)
            elif self.kind == 'cs':
                return -jacobi_ds(x, m) * jacobi_ns(x, m)
            elif self.kind == 'dc':
                return (Integer(1) - m) * jacobi_nc(x, m) * jacobi_sc(x, m)
            elif self.kind == 'dn':
                return -m * jacobi_sn(x, m) * jacobi_cn(x, m)
            elif self.kind == 'ds':
                return -jacobi_cs(x, m) * jacobi_ns(x, m)
            elif self.kind == 'nc':
                return jacobi_dc(x, m) * jacobi_sc(x, m)
            elif self.kind == 'nd':
                return m * jacobi_cd(x, m) * jacobi_sd(x, m)
            elif self.kind == 'ns':
                return -jacobi_cs(x, m) * jacobi_ds(x, m)
            elif self.kind == 'sc':
                return jacobi_dc(x, m) * jacobi_nc(x, m)
            elif self.kind == 'sd':
                return jacobi_cd(x, m) * jacobi_nd(x, m)
            elif self.kind == 'sn':
                return jacobi_cn(x, m) * jacobi_dn(x, m)
        elif diff_param == 1:
            # From Maxima
            if self.kind == 'nd':
                return (HALF*((x + elliptic_e(arcsin(jacobi_sn(x, m)), m) /
                              (m - Integer(1)))*jacobi_sn(x, m)*jacobi_cn(x, m) -
                              jacobi_dn(x, m)*jacobi_sn(x, m)**Integer(2)/(m - Integer(1))) /
                        jacobi_dn(x, m)**Integer(2))
            elif self.kind == 'ns':
                return (HALF*(jacobi_sn(x, m)*jacobi_cn(x, m)**Integer(2)/(m - Integer(1)) -
                              (x + elliptic_e(arcsin(jacobi_sn(x, m)), m) /
                               (m - Integer(1)))*jacobi_dn(x, m)*jacobi_cn(x, m)/m) /
                        jacobi_sn(x, m)**Integer(2))
            elif self.kind == 'nc':
                return (-HALF*(jacobi_sn(x, m)**Integer(2)*jacobi_cn(x, m)/(m - Integer(1)) -
                               (x + elliptic_e(arcsin(jacobi_sn(x, m)), m) /
                                (m - Integer(1)))*jacobi_dn(x, m) *
                               jacobi_sn(x, m)/m)/jacobi_cn(x, m)**Integer(2))
            elif self.kind == 'dn':
                return (-HALF*(x + elliptic_e(arcsin(jacobi_sn(x, m)), m) /
                               (m - Integer(1)))*jacobi_sn(x, m)*jacobi_cn(x, m) +
                        HALF*jacobi_dn(x, m)*jacobi_sn(x, m)**Integer(2)/(m - Integer(1)))
            elif self.kind == 'ds':
                return (HALF*(jacobi_sn(x, m)*jacobi_cn(x, m)**Integer(2)/(m - Integer(1)) -
                        (x + elliptic_e(arcsin(jacobi_sn(x, m)), m) /
                         (m - Integer(1)))*jacobi_dn(x, m)*jacobi_cn(x, m)/m) *
                        jacobi_dn(x, m)/jacobi_sn(x, m)**Integer(2) -
                        HALF*((x + elliptic_e(arcsin(jacobi_sn(x, m)), m) /
                              (m - Integer(1)))*jacobi_sn(x, m)*jacobi_cn(x, m) -
                              jacobi_dn(x, m)*jacobi_sn(x, m)**Integer(2)/(m - Integer(1))) /
                        jacobi_sn(x, m))
            elif self.kind == 'dc':
                return (-HALF*(jacobi_sn(x, m)**Integer(2)*jacobi_cn(x, m)/(m - Integer(1)) -
                               (x + elliptic_e(arcsin(jacobi_sn(x, m)), m) /
                                (m - Integer(1)))*jacobi_dn(x, m) *
                               jacobi_sn(x, m)/m)*jacobi_dn(x, m) /
                        jacobi_cn(x, m)**Integer(2) -
                        HALF*((x + elliptic_e(arcsin(jacobi_sn(x, m)), m) /
                               (m - Integer(1)))*jacobi_sn(x, m)*jacobi_cn(x, m) -
                              jacobi_dn(x, m)*jacobi_sn(x, m)**Integer(2)/(m - Integer(1))) /
                        jacobi_cn(x, m))
            elif self.kind == 'sn':
                return (-HALF*jacobi_sn(x, m)*jacobi_cn(x, m)**Integer(2)/(m - Integer(1)) +
                        HALF*(x + elliptic_e(arcsin(jacobi_sn(x, m)), m) /
                              (m - Integer(1)))*jacobi_dn(x, m)*jacobi_cn(x, m)/m)
            elif self.kind == 'sd':
                return (-HALF*(jacobi_sn(x, m)*jacobi_cn(x, m)**Integer(2)/(m - Integer(1)) -
                        (x + elliptic_e(arcsin(jacobi_sn(x, m)), m) /
                         (m - Integer(1)))*jacobi_dn(x, m)*jacobi_cn(x, m)/m) /
                        jacobi_dn(x, m) + HALF *
                        ((x + elliptic_e(arcsin(jacobi_sn(x, m)), m) /
                          (m - Integer(1)))*jacobi_sn(x, m)*jacobi_cn(x, m) -
                         jacobi_dn(x, m)*jacobi_sn(x, m)**Integer(2)/(m - Integer(1))) *
                        jacobi_sn(x, m)/jacobi_dn(x, m)**Integer(2))
            elif self.kind == 'sc':
                return (-HALF*(jacobi_sn(x, m)*jacobi_cn(x, m)**Integer(2)/(m - Integer(1)) -
                               (x + elliptic_e(arcsin(jacobi_sn(x, m)), m) /
                                (m - Integer(1)))*jacobi_dn(x, m) *
                               jacobi_cn(x, m)/m)/jacobi_cn(x, m) -
                        HALF*(jacobi_sn(x, m)**Integer(2)*jacobi_cn(x, m)/(m - Integer(1)) -
                              (x + elliptic_e(arcsin(jacobi_sn(x, m)), m) /
                               (m - Integer(1)))*jacobi_dn(x, m)*jacobi_sn(x, m)/m) *
                        jacobi_sn(x, m)/jacobi_cn(x, m)**Integer(2))
            elif self.kind == 'cn':
                return (HALF*jacobi_sn(x, m)**Integer(2)*jacobi_cn(x, m)/(m - Integer(1)) -
                        HALF*(x + elliptic_e(arcsin(jacobi_sn(x, m)), m) /
                              (m - Integer(1)))*jacobi_dn(x, m)*jacobi_sn(x, m)/m)
            elif self.kind == 'cd':
                return (HALF*(jacobi_sn(x, m)**Integer(2)*jacobi_cn(x, m)/(m - Integer(1)) -
                        (x + elliptic_e(arcsin(jacobi_sn(x, m)), m) /
                         (m - Integer(1)))*jacobi_dn(x, m)*jacobi_sn(x, m)/m) /
                        jacobi_dn(x, m) +
                        HALF*((x + elliptic_e(arcsin(jacobi_sn(x, m)), m) /
                               (m - Integer(1)))*jacobi_sn(x, m)*jacobi_cn(x, m) -
                              jacobi_dn(x, m)*jacobi_sn(x, m)**Integer(2)/(m - Integer(1))) *
                        jacobi_cn(x, m)/jacobi_dn(x, m)**Integer(2))
            elif self.kind == 'cs':
                return (HALF*(jacobi_sn(x, m)*jacobi_cn(x, m)**Integer(2)/(m - Integer(1)) -
                        (x + elliptic_e(arcsin(jacobi_sn(x, m)), m) /
                         (m - Integer(1)))*jacobi_dn(x, m)*jacobi_cn(x, m)/m) *
                        jacobi_cn(x, m)/jacobi_sn(x, m)**Integer(2) +
                        HALF*(jacobi_sn(x, m)**Integer(2)*jacobi_cn(x, m)/(m - Integer(1)) -
                              (x + elliptic_e(arcsin(jacobi_sn(x, m)), m) /
                               (m - Integer(1)))*jacobi_dn(x, m)*jacobi_sn(x, m)/m) /
                        jacobi_sn(x, m))

    def _latex_(self):
        r"""
        TESTS::

            sage: latex(jacobi_sn)
            \operatorname{sn}
        """
        return r"\operatorname{{{}}}".format(self.kind)

    def _print_latex_(self, x, m):
        r"""
        TESTS::

            sage: latex(jacobi_sn(x, 3))                                                # needs sage.symbolic
            \operatorname{sn}\left(x\middle|3\right)
        """
        return r"\operatorname{{{}}}\left({}\middle|{}\right)".format(self.kind,
                                                                      latex(x),
                                                                      latex(m))


jacobi_nd = Jacobi('nd')
jacobi_ns = Jacobi('ns')
jacobi_nc = Jacobi('nc')
jacobi_dn = Jacobi('dn')
jacobi_ds = Jacobi('ds')
jacobi_dc = Jacobi('dc')
jacobi_sn = Jacobi('sn')
jacobi_sd = Jacobi('sd')
jacobi_sc = Jacobi('sc')
jacobi_cn = Jacobi('cn')
jacobi_cd = Jacobi('cd')
jacobi_cs = Jacobi('cs')


class InverseJacobi(BuiltinFunction):
    r"""
    Base class for the inverse Jacobi elliptic functions.
    """
    def __init__(self, kind):
        r"""
        Initialize ``self``.

        EXAMPLES::

            sage: from sage.functions.jacobi import InverseJacobi
            sage: InverseJacobi('sn')
            inverse_jacobi_sn
        """
        if kind not in ['nd', 'ns', 'nc', 'dn', 'ds', 'dc', 'sn', 'sd',
                        'sc', 'cn', 'cd', 'cs']:
            raise ValueError("kind must be one of 'nd', 'ns', 'nc', 'dn', "
                             "'ds', 'dc', 'sn', 'sd', 'sc', 'cn', 'cd', 'cs'.")
        self.kind = kind
        BuiltinFunction.__init__(self,
            name=f'inverse_jacobi_{kind}', nargs=2, evalf_params_first=False,
            conversions=dict(maple=('InverseJacobi{}'.format(kind.upper())),
                             mathematica=f'InverseJacobi{kind.upper()}',
                             maxima=(f'inverse_jacobi_{kind}')))

    def _eval_(self, x, m):
        r"""
        TESTS:

        Check that the simplifications are correct::

            sage: # needs mpmath sage.symbolic
            sage: from mpmath import almosteq
            sage: almosteq(n(inverse_jacobi_cd(1, -8, hold=True)),
            ....:          n(inverse_jacobi_cd(1, -8)))
            True
            sage: almosteq(n(inverse_jacobi_cn(0, -5, hold=True)),
            ....:          n(inverse_jacobi_cn(0, -5)))
            True
            sage: almosteq(n(inverse_jacobi_cn(1, -8, hold=True)),
            ....:          n(inverse_jacobi_cn(1, -8)))
            True
            sage: almosteq(n(inverse_jacobi_cs(7, 1, hold=True)),
            ....:          n(inverse_jacobi_cs(7, 1)))
            True
            sage: almosteq(n(inverse_jacobi_dc(3, 0, hold=True)),
            ....:          n(inverse_jacobi_dc(3, 0)))
            True
            sage: almosteq(n(inverse_jacobi_dc(1, 7, hold=True)),
            ....:          n(inverse_jacobi_dc(1, 7)))
            True
            sage: almosteq(n(inverse_jacobi_dn(1, -1, hold=True)),
            ....:          n(inverse_jacobi_dn(1, -1)))
            True
            sage: almosteq(n(inverse_jacobi_ds(7, 0, hold=True)),
            ....:          n(inverse_jacobi_ds(7, 0)))
            True
            sage: almosteq(n(inverse_jacobi_ds(5, 1, hold=True)),
            ....:          n(inverse_jacobi_ds(5, 1)))
            True
            sage: almosteq(n(inverse_jacobi_nc(-2, 0, hold=True)),
            ....:          n(inverse_jacobi_nc(-2, 0)))
            True
            sage: almosteq(n(inverse_jacobi_nc(-1, 1, hold=True)),
            ....:          n(inverse_jacobi_nc(-1, 1)))
            True
            sage: almosteq(n(inverse_jacobi_nc(1, 4, hold=True)),
            ....:          n(inverse_jacobi_nc(1, 4)))
            True
            sage: almosteq(n(inverse_jacobi_nd(9, 1, hold=True)),
            ....:          n(inverse_jacobi_nd(9, 1)))
            True
            sage: almosteq(n(inverse_jacobi_nd(1, -9, hold=True)),
            ....:          n(inverse_jacobi_nd(1, -9)))
            True
            sage: almosteq(n(inverse_jacobi_ns(-6, 0, hold=True)),
            ....:          n(inverse_jacobi_ns(-6, 0)))
            True
            sage: almosteq(n(inverse_jacobi_ns(6, 1, hold=True)),
            ....:          n(inverse_jacobi_ns(6, 1)))
            True
            sage: almosteq(n(inverse_jacobi_sc(9, 0, hold=True)),
            ....:          n(inverse_jacobi_sc(9, 0)))
            True
            sage: almosteq(n(inverse_jacobi_sc(8, 1, hold=True)),
            ....:          n(inverse_jacobi_sc(8, 1)))
            True
            sage: almosteq(n(inverse_jacobi_sc(0, -8, hold=True)),
            ....:          n(inverse_jacobi_sc(0, -8)))
            True
            sage: almosteq(n(inverse_jacobi_sd(-1, 0, hold=True)),
            ....:          n(inverse_jacobi_sd(-1, 0)))
            True
            sage: almosteq(n(inverse_jacobi_sd(-2, 1, hold=True)),
            ....:          n(inverse_jacobi_sd(-2, 1)))
            True
            sage: almosteq(n(inverse_jacobi_sd(0, -2, hold=True)),
            ....:          n(inverse_jacobi_sd(0, -2)))
            True
            sage: almosteq(n(inverse_jacobi_sn(0, 0, hold=True)),
            ....:          n(inverse_jacobi_sn(0, 0)))
            True
            sage: almosteq(n(inverse_jacobi_sn(0, 6, hold=True)),
            ....:          n(inverse_jacobi_sn(0, 6)))
            True
        """
        if self.kind == 'cd':
            if m == 0:
                return arccos(x)
            elif x == 1:
                return Integer(0)
        elif self.kind == 'cn':
            if m == 0:
                return arccos(x)
            elif m == 1:
                return arcsech(x)
            elif x == 0:
                return elliptic_kc(m)
            elif x == 1:
                return Integer(0)
        elif self.kind == 'cs':
            if m == 0:
                return arccot(x)
            elif m == 1:
                return arccsch(x)
        elif self.kind == 'dc':
            if m == 0:
                return arcsec(x)
            elif x == 1:
                return Integer(0)
        elif self.kind == 'dn':
            if m == 1:
                return arcsech(x)
            elif x == 1:
                return Integer(0)
        elif self.kind == 'ds':
            if m == 0:
                return arccsc(x)
            elif m == 1:
                return arccsch(x)
        elif self.kind == 'nc':
            if m == 0:
                return arcsec(x)
            elif m == 1:
                return arccosh(x)
            elif x == 1:
                return Integer(0)
        elif self.kind == 'nd':
            if m == 1:
                return arccosh(x)
            elif x == 1:
                return Integer(0)
        elif self.kind == 'ns':
            if m == 0:
                return arccsc(x)
            elif m == 1:
                return arccoth(x)
        elif self.kind == 'sc':
            if m == 0:
                return arctan(x)
            elif m == 1:
                return arcsinh(x)
            elif x == 0:
                return Integer(0)
        elif self.kind == 'sd':
            if m == 0:
                return arcsin(x)
            elif m == 1:
                return arcsinh(x)
            elif x == 0:
                return Integer(0)
        elif self.kind == 'sn':
            if m == 0:
                return arcsin(x)
            elif m == 1:
                return arctanh(x)
            elif x == 0:
                return Integer(0)
        return

    def _evalf_(self, x, m, parent=None, algorithm=None):
        r"""
        TESTS::

            sage: inverse_jacobi_cn(2, 3).n()                                           # needs mpmath
            0.859663746362987*I
            sage: inverse_jacobi_cd(3, 4).n(100)                                        # needs mpmath
            -0.67214752201235862490069823239 + 2.1565156474996432354386749988*I
        """
        return _mpmath_utils_call(inverse_jacobi_f, self.kind, x, m, parent=parent)

    def _derivative_(self, x, m, diff_param):
        r"""
        TESTS:

        Check that ``dy/dx * dx/dy == 1``, where ``y = jacobi_pq(x, m)`` and
        ``x = inverse_jacobi_pq(y, m)``::

            sage: # needs mpmath sage.symbolic
            sage: from mpmath import almosteq
            sage: a = 0.130103220857094
            sage: b = 0.437176765041986
            sage: m = var('m')
            sage: almosteq(abs((diff(jacobi_cd(x, m), x) *
            ....:               diff(inverse_jacobi_cd(x, m), x).subs(x=jacobi_cd(x, m))).subs(x=a, m=b)),
            ....:          1, abs_eps=1e-14)
            True
            sage: almosteq(abs((diff(jacobi_cn(x, m), x) *
            ....:               diff(inverse_jacobi_cn(x, m), x).subs(x=jacobi_cn(x, m))).subs(x=a, m=b)),
            ....:          1, abs_eps=1e-14)
            True
            sage: almosteq(abs((diff(jacobi_cs(x, m), x) *
            ....:               diff(inverse_jacobi_cs(x, m), x).subs(x=jacobi_cs(x, m))).subs(x=a, m=b)),
            ....:          1, abs_eps=1e-14)
            True
            sage: almosteq(abs((diff(jacobi_dc(x, m), x) *
            ....:               diff(inverse_jacobi_dc(x, m), x).subs(x=jacobi_dc(x, m))).subs(x=a, m=b)),
            ....:          1, abs_eps=1e-14)
            True
            sage: almosteq(abs((diff(jacobi_dn(x, m), x) *
            ....:               diff(inverse_jacobi_dn(x, m), x).subs(x=jacobi_dn(x, m))).subs(x=a, m=b)),
            ....:          1, abs_eps=1e-14)
            True
            sage: almosteq(abs((diff(jacobi_ds(x, m), x) *
            ....:               diff(inverse_jacobi_ds(x, m), x).subs(x=jacobi_ds(x, m))).subs(x=a, m=b)),
            ....:          1, abs_eps=1e-14)
            True
            sage: almosteq(abs((diff(jacobi_nc(x, m), x) *
            ....:               diff(inverse_jacobi_nc(x, m), x).subs(x=jacobi_nc(x, m))).subs(x=a, m=b)),
            ....:          1, abs_eps=1e-14)
            True
            sage: almosteq(abs((diff(jacobi_nd(x, m), x) *
            ....:               diff(inverse_jacobi_nd(x, m), x).subs(x=jacobi_nd(x, m))).subs(x=a, m=b)),
            ....:          1, abs_eps=1e-14)
            True
            sage: almosteq(abs((diff(jacobi_ns(x, m), x) *
            ....:               diff(inverse_jacobi_ns(x, m), x).subs(x=jacobi_ns(x, m))).subs(x=a, m=b)),
            ....:          1, abs_eps=1e-14)
            True
            sage: almosteq(abs((diff(jacobi_sc(x, m), x) *
            ....:               diff(inverse_jacobi_sc(x, m), x).subs(x=jacobi_sc(x, m))).subs(x=a, m=b)),
            ....:          1, abs_eps=1e-14)
            True
            sage: almosteq(abs((diff(jacobi_sd(x, m), x) *
            ....:               diff(inverse_jacobi_sd(x, m), x).subs(x=jacobi_sd(x, m))).subs(x=a, m=b)),
            ....:          1, abs_eps=1e-14)
            True
            sage: almosteq(abs((diff(jacobi_sn(x, m), x) *
            ....:               diff(inverse_jacobi_sn(x, m), x).subs(x=jacobi_sn(x, m))).subs(x=a, m=b)),
            ....:          1, abs_eps=1e-14)
            True
        """
        # From Wolfram Functions Site
        if diff_param == 0:
            if self.kind == 'cd':
                return (jacobi_sn(inverse_jacobi_cd(x, m), m) /
                        (x ** Integer(2) - Integer(1)))
            elif self.kind == 'cn':
                return (jacobi_ds(inverse_jacobi_cn(x, m), m) /
                        (m * x ** Integer(2) - m + Integer(1)))
            elif self.kind == 'cs':
                return (jacobi_nd(inverse_jacobi_cs(x, m), m) /
                        (x ** Integer(2) + Integer(1)))
            elif self.kind == 'dc':
                return (jacobi_sn(inverse_jacobi_dc(x, m), m) /
                        (x ** Integer(2) - Integer(1)))
            elif self.kind == 'dn':
                return -(jacobi_cs(inverse_jacobi_dn(x, m), m) /
                         (x ** Integer(2) + m - Integer(1)))
            elif self.kind == 'ds':
                return (jacobi_nc(inverse_jacobi_ds(x, m), m) /
                        (x ** Integer(2) + m))
            elif self.kind == 'nc':
                return (jacobi_ds(inverse_jacobi_nc(x, m), m) /
                        (-m * x ** Integer(2) + x ** Integer(2) + m))
            elif self.kind == 'nd':
                return (jacobi_sc(inverse_jacobi_nd(x, m), m) /
                        (x ** Integer(2) - Integer(1)))
            elif self.kind == 'ns':
                return Integer(1) / (jacobi_cs(inverse_jacobi_ns(x, m), m) *
                                     jacobi_ds(inverse_jacobi_ns(x, m), m))
            elif self.kind == 'sc':
                return (jacobi_nd(inverse_jacobi_sc(x, m), m) /
                        (x ** Integer(2) + Integer(1)))
            elif self.kind == 'sd':
                return (jacobi_cn(inverse_jacobi_sd(x, m), m) /
                        ((m - Integer(1)) * x ** Integer(2) + Integer(1)))
            elif self.kind == 'sn':
                return (jacobi_cd(inverse_jacobi_sn(x, m), m) /
                        (Integer(1) - x ** Integer(2)))
        elif diff_param == 1:
            if self.kind == 'cd':
                return ((Integer(1) / (Integer(2) * (Integer(1) - m) * m)) *
                        ((m - Integer(1)) * inverse_jacobi_cd(x, m) +
                         elliptic_e(jacobi_am(inverse_jacobi_cd(x, m), m),
                                    m)))
            elif self.kind == 'cn':
                return ((-(Integer(1) / (Integer(2) * (-Integer(1) + m) * m))) *
                        (elliptic_e(jacobi_am(inverse_jacobi_cn(x, m), m),
                                    m) + (-Integer(1) + m) *
                         inverse_jacobi_cn(x, m) - m * x *
                         jacobi_sd(inverse_jacobi_cn(x, m), m)))
            elif self.kind == 'cs':
                return ((-(Integer(1) / (Integer(2) * (-Integer(1) + m) * m * (Integer(1) + x ** Integer(2))))) *
                        ((Integer(1) + x ** Integer(2)) *
                         elliptic_e(jacobi_am(inverse_jacobi_cs(x, m), m),
                                    m) + (-Integer(1) + m) * (Integer(1) + x ** Integer(2)) *
                         inverse_jacobi_cs(x, m) - m * x *
                         jacobi_nd(inverse_jacobi_cs(x, m), m)))
            elif self.kind == 'dc':
                return ((Integer(1) / (Integer(2) * (Integer(1) - m) * m)) *
                        (elliptic_e(jacobi_am(inverse_jacobi_dc(x, m), m),
                                    m) - (Integer(1) - m) *
                         inverse_jacobi_dc(x, m)))
            elif self.kind == 'dn':
                return ((Integer(1) / (Integer(2) * (Integer(1) - m) * m)) * ((m - Integer(1)) *
                        inverse_jacobi_dn(x, m) +
                        elliptic_e(jacobi_am(inverse_jacobi_dn(x, m), m), m) -
                        x * jacobi_sc(inverse_jacobi_dn(x, m), m)))
            elif self.kind == 'ds':
                return ((-(Integer(1) / (Integer(2) * (-Integer(1) + m) * m))) *
                        (elliptic_e(jacobi_am(inverse_jacobi_ds(x, m), m), m) +
                         (-Integer(1) + m) * inverse_jacobi_ds(x, m) -
                         (m * x * jacobi_nc(inverse_jacobi_ds(x, m), m)) /
                         (m + x ** Integer(2))))
            elif self.kind == 'nc':
                return ((Integer(1) / (Integer(2) * (-Integer(1) + m) * m * x)) * ((-x) *
                        (elliptic_e(jacobi_am(inverse_jacobi_nc(x, m), m), m) +
                         (-Integer(1) + m) * inverse_jacobi_nc(x, m)) + m *
                        jacobi_sd(inverse_jacobi_nc(x, m), m)))
            elif self.kind == 'nd':
                return ((Integer(1) / (Integer(2) * (m - Integer(1)) * m)) *
                        ((Integer(1) - m) * inverse_jacobi_nd(x, m) -
                         elliptic_e(jacobi_am(inverse_jacobi_nd(x, m), m), m) +
                         (Integer(1) / x) * jacobi_sc(inverse_jacobi_nd(x, m), m)))
            elif self.kind == 'ns':
                return ((Integer(1)/(Integer(2) * (m - Integer(1)) * m)) *
                        ((Integer(1) - m) * inverse_jacobi_ns(x, m) -
                         elliptic_e(jacobi_am(inverse_jacobi_ns(x, m), m), m) +
                         (m / x) * jacobi_cd(inverse_jacobi_ns(x, m), m)))
            elif self.kind == 'sc':
                return ((-(Integer(1) / (Integer(2) * (-Integer(1) + m) * m * (Integer(1) + x ** Integer(2))))) *
                        ((Integer(1) + x ** Integer(2)) *
                         elliptic_e(jacobi_am(inverse_jacobi_sc(x, m), m), m) +
                         (-Integer(1) + m) * (Integer(1) + x ** Integer(2)) * inverse_jacobi_sc(x, m) -
                         m * x * jacobi_nd(inverse_jacobi_sc(x, m), m)))
            elif self.kind == 'sd':
                return ((-(Integer(1) / (Integer(2) * (-Integer(1) + m) * m))) *
                        (elliptic_e(jacobi_am(inverse_jacobi_sd(x, m), m), m) +
                         (-Integer(1) + m) * inverse_jacobi_sd(x, m) -
                         (m * x * jacobi_nc(inverse_jacobi_sd(x, m), m)) /
                         (Integer(1) + m * x ** Integer(2))))
            elif self.kind == 'sn':
                return ((Integer(1) / (Integer(2) * (Integer(1) - m) * m)) *
                        (elliptic_e(jacobi_am(inverse_jacobi_sn(x, m), m), m) +
                         (-Integer(1) + m) * inverse_jacobi_sn(x, m) - m * x *
                         jacobi_cd(inverse_jacobi_sn(x, m), m)))

    def _latex_(self):
        r"""
        TESTS::

            sage: latex(inverse_jacobi_dn)
            \operatorname{arcdn}
        """
        return r"\operatorname{{arc{}}}".format(self.kind)

    def _print_latex_(self, x, m):
        r"""
        TESTS::

            sage: latex(inverse_jacobi_dn(x, 3))                                        # needs sage.symbolic
            \operatorname{arcdn}\left(x\middle|3\right)
        """
        return r"\operatorname{{arc{}}}\left({}\middle|{}\right)".format(self.kind,
                                                                         latex(x),
                                                                         latex(m))


inverse_jacobi_nd = InverseJacobi('nd')
inverse_jacobi_ns = InverseJacobi('ns')
inverse_jacobi_nc = InverseJacobi('nc')
inverse_jacobi_dn = InverseJacobi('dn')
inverse_jacobi_ds = InverseJacobi('ds')
inverse_jacobi_dc = InverseJacobi('dc')
inverse_jacobi_sn = InverseJacobi('sn')
inverse_jacobi_sd = InverseJacobi('sd')
inverse_jacobi_sc = InverseJacobi('sc')
inverse_jacobi_cn = InverseJacobi('cn')
inverse_jacobi_cd = InverseJacobi('cd')
inverse_jacobi_cs = InverseJacobi('cs')


def jacobi(kind, z, m, **kwargs):
    r"""
    The 12 Jacobi elliptic functions.

    INPUT:

    - ``kind`` -- string of the form ``'pq'``, where ``p``, ``q`` are in
      ``c``, ``d``, ``n``, ``s``
    - ``z`` -- a complex number
    - ``m`` -- a complex number; note that `m = k^2`, where `k` is
      the elliptic modulus

    EXAMPLES::

        sage: # needs mpmath
        sage: jacobi('sn', 1, 1)
        tanh(1)
        sage: jacobi('cd', 1, 1/2)
        jacobi_cd(1, 1/2)
        sage: RDF(jacobi('cd', 1, 1/2))
        0.7240097216593705
        sage: (RDF(jacobi('cn', 1, 1/2)), RDF(jacobi('dn', 1, 1/2)),
        ....:  RDF(jacobi('cn', 1, 1/2) / jacobi('dn', 1, 1/2)))
        (0.5959765676721407, 0.8231610016315962, 0.7240097216593705)

        sage: jsn = jacobi('sn', x, 1)                                                  # needs sage.symbolic
        sage: P = plot(jsn, 0, 1)                                                       # needs sage.plot sage.symbolic
    """
    if kind == 'nd':
        return jacobi_nd(z, m, **kwargs)
    elif kind == 'ns':
        return jacobi_ns(z, m, **kwargs)
    elif kind == 'nc':
        return jacobi_nc(z, m, **kwargs)
    elif kind == 'dn':
        return jacobi_dn(z, m, **kwargs)
    elif kind == 'ds':
        return jacobi_ds(z, m, **kwargs)
    elif kind == 'dc':
        return jacobi_dc(z, m, **kwargs)
    elif kind == 'sn':
        return jacobi_sn(z, m, **kwargs)
    elif kind == 'sd':
        return jacobi_sd(z, m, **kwargs)
    elif kind == 'sc':
        return jacobi_sc(z, m, **kwargs)
    elif kind == 'cn':
        return jacobi_cn(z, m, **kwargs)
    elif kind == 'cd':
        return jacobi_cd(z, m, **kwargs)
    elif kind == 'cs':
        return jacobi_cs(z, m, **kwargs)
    else:
        raise ValueError("kind must be one of 'nd', 'ns', 'nc', 'dn', "
                         "'ds', 'dc', 'sn', 'sd', 'sc', 'cn', 'cd', 'cs'.")


def inverse_jacobi(kind, x, m, **kwargs):
    r"""
    The inverses of the 12 Jacobi elliptic functions. They have the property
    that

    .. MATH::

        \operatorname{pq}(\operatorname{arcpq}(x|m)|m) =
        \operatorname{pq}(\operatorname{pq}^{-1}(x|m)|m) = x.

    INPUT:

    - ``kind`` -- string of the form ``'pq'``, where ``p``, ``q`` are in
      ``c``, ``d``, ``n``, ``s``
    - ``x`` -- a real number
    - ``m`` -- a real number; note that `m = k^2`, where `k` is the elliptic
      modulus

    EXAMPLES::

        sage: jacobi('dn', inverse_jacobi('dn', 3, 0.4), 0.4)                           # needs mpmath
        3.00000000000000
        sage: inverse_jacobi('dn', 10, 1/10).n(digits=50)                               # needs mpmath
        2.4777736267904273296523691232988240759001423661683*I
        sage: inverse_jacobi_dn(x, 1)                                                   # needs sage.symbolic
        arcsech(x)
        sage: inverse_jacobi_dn(1, 3)                                                   # needs mpmath
        0

        sage: # needs sage.symbolic
        sage: m = var('m')
        sage: z = inverse_jacobi_dn(x, m).series(x, 4).subs(x=0.1, m=0.7)
        sage: jacobi_dn(z, 0.7)
        0.0999892750039819...
        sage: inverse_jacobi_nd(x, 1)
        arccosh(x)

        sage: # needs mpmath
        sage: inverse_jacobi_nd(1, 2)
        0
        sage: inverse_jacobi_ns(10^-5, 3).n()
        5.77350269202456e-6 + 1.17142008414677*I
        sage: jacobi('sn', 1/2, 1/2)
        jacobi_sn(1/2, 1/2)
        sage: jacobi('sn', 1/2, 1/2).n()
        0.470750473655657
        sage: inverse_jacobi('sn', 0.47, 1/2)
        0.499098231322220
        sage: inverse_jacobi('sn', 0.4707504, 0.5)
        0.499999911466555
        sage: P = plot(inverse_jacobi('sn', x, 0.5), 0, 1)                              # needs sage.plot
    """
    if kind == 'nd':
        return inverse_jacobi_nd(x, m, **kwargs)
    elif kind == 'ns':
        return inverse_jacobi_ns(x, m, **kwargs)
    elif kind == 'nc':
        return inverse_jacobi_nc(x, m, **kwargs)
    elif kind == 'dn':
        return inverse_jacobi_dn(x, m, **kwargs)
    elif kind == 'ds':
        return inverse_jacobi_ds(x, m, **kwargs)
    elif kind == 'dc':
        return inverse_jacobi_dc(x, m, **kwargs)
    elif kind == 'sn':
        return inverse_jacobi_sn(x, m, **kwargs)
    elif kind == 'sd':
        return inverse_jacobi_sd(x, m, **kwargs)
    elif kind == 'sc':
        return inverse_jacobi_sc(x, m, **kwargs)
    elif kind == 'cn':
        return inverse_jacobi_cn(x, m, **kwargs)
    elif kind == 'cd':
        return inverse_jacobi_cd(x, m, **kwargs)
    elif kind == 'cs':
        return inverse_jacobi_cs(x, m, **kwargs)
    else:
        raise ValueError("kind must be one of 'nd', 'ns', 'nc', 'dn', "
                         "'ds', 'dc', 'sn', 'sd', 'sc', 'cn', 'cd', 'cs'.")


class JacobiAmplitude(BuiltinFunction):
    r"""
    The Jacobi amplitude function
    `\operatorname{am}(x|m) = \int_0^x \operatorname{dn}(t|m) dt` for
    `-K(m) \leq x \leq K(m)`, `F(\operatorname{am}(x|m)|m) = x`.
    """
    def __init__(self):
        r"""
        TESTS::

            sage: from sage.functions.jacobi import JacobiAmplitude
            sage: JacobiAmplitude()
            jacobi_am
        """
        BuiltinFunction.__init__(self, name='jacobi_am', nargs=2,
            conversions=dict(maple='JacobiAM',
                             mathematica='JacobiAmplitude'),
            evalf_params_first=False)

    def _eval_(self, x, m):
        r"""
        TESTS::

            sage: jacobi_am(x, 0)                                                       # needs sage.symbolic
            x
            sage: jacobi_am(0, x)                                                       # needs sage.symbolic
            0
            sage: jacobi_am(3, 4.)                                                      # needs mpmath
            -0.339059208303591
        """
        if m == 0:
            return x
        elif x == 0:
            return Integer(0)
        return

    def _evalf_(self, x, m, parent=None, algorithm=None):
        r"""
        TESTS::

            sage: jacobi_am(1, 2).n(100)                                                # needs mpmath
            0.73704379494724574105101929735
        """
        return _mpmath_utils_call(jacobi_am_f, x, m, parent=parent)

    def _derivative_(self, x, m, diff_param):
        r"""
        TESTS::

            sage: diff(jacobi_am(x, 3), x)                                              # needs sage.symbolic
            jacobi_dn(x, 3)
            sage: diff(jacobi_am(3, x), x)                                              # needs sage.symbolic
            -1/2*(x*jacobi_cn(3, x)*jacobi_sn(3, x) -...
            (3*x + elliptic_e(jacobi_am(3, x), x) - 3)*jacobi_dn(3, x))/((x - 1)*x)
        """
        if diff_param == 0:
            return jacobi_dn(x, m)
        elif diff_param == 1:
            return (((Integer(-1) + m) * x + elliptic_e(jacobi_am(x, m), m)) *
                    jacobi('dn', x, m) - m * jacobi('cn', x, m) *
                    jacobi('sn', x, m)) / (Integer(2) * (Integer(-1) + m) * m)

    def _latex_(self):
        r"""
        TESTS::

            sage: latex(jacobi_am)
            \operatorname{am}
        """
        return r"\operatorname{am}"

    def _print_latex_(self, x, m):
        r"""
        TESTS::

            sage: latex(jacobi_am(3,x))                                                 # needs sage.symbolic
            \operatorname{am}\left(3\middle|x\right)
        """
        return r"\operatorname{{am}}\left({}\middle|{}\right)".format(latex(x),
                                                                      latex(m))


jacobi_am = JacobiAmplitude()


def inverse_jacobi_f(kind, x, m):
    r"""
    Internal function for numerical evaluation of a continuous complex branch
    of each inverse Jacobi function, as described in [Tee1997]_. Only accepts
    real arguments.

    TESTS::

        sage: from mpmath import ellipfun, chop                                         # needs mpmath
        sage: from sage.functions.jacobi import inverse_jacobi_f

        sage: # needs mpmath
        sage: chop(ellipfun('sn', inverse_jacobi_f('sn', 0.6, 0), 0))
        mpf('0.59999999999999998')
        sage: chop(ellipfun('sn', inverse_jacobi_f('sn', 0.6, 1), 1))
        mpf('0.59999999999999998')
        sage: chop(ellipfun('sn', inverse_jacobi_f('sn', 0, -3), -3))
        mpf('0.0')
        sage: chop(ellipfun('sn', inverse_jacobi_f('sn', -1, 4), 4))
        mpf('-1.0')
        sage: chop(ellipfun('sn', inverse_jacobi_f('sn', 0.3, 4), 4))
        mpf('0.29999999999999999')
        sage: chop(ellipfun('sn', inverse_jacobi_f('sn', 0.8, 4), 4))
        mpf('0.80000000000000004')

        sage: # needs mpmath
        sage: chop(ellipfun('ns', inverse_jacobi_f('ns', 0.8, 0), 0))
        mpf('0.80000000000000004')
        sage: chop(ellipfun('ns', inverse_jacobi_f('ns', -0.7, 1), 1))
        mpf('-0.69999999999999996')
        sage: chop(ellipfun('ns', inverse_jacobi_f('ns', 0.01, 2), 2))
        mpf('0.01')
        sage: chop(ellipfun('ns', inverse_jacobi_f('ns', 0, 2), 2))
        mpf('0.0')
        sage: chop(ellipfun('ns', inverse_jacobi_f('ns', -10, 6), 6))
        mpf('-10.0')

        sage: # needs mpmath
        sage: chop(ellipfun('cn', inverse_jacobi_f('cn', -10, 0), 0))
        mpf('-9.9999999999999982')
        sage: chop(ellipfun('cn', inverse_jacobi_f('cn', 50, 1), 1))
        mpf('50.000000000000071')
        sage: chop(ellipfun('cn', inverse_jacobi_f('cn', 1, 5), 5))
        mpf('1.0')
        sage: chop(ellipfun('cn', inverse_jacobi_f('cn', 0.5, -5), -5))
        mpf('0.5')
        sage: chop(ellipfun('cn', inverse_jacobi_f('cn', -0.75, -15), -15))
        mpf('-0.75000000000000022')
        sage: chop(ellipfun('cn', inverse_jacobi_f('cn', 10, 0.8), 0.8))
        mpf('9.9999999999999982')
        sage: chop(ellipfun('cn', inverse_jacobi_f('cn', -2, 0.9), 0.9))
        mpf('-2.0')

        sage: # needs mpmath
        sage: chop(ellipfun('nc', inverse_jacobi_f('nc', -4, 0), 0))
        mpf('-3.9999999999999987')
        sage: chop(ellipfun('nc', inverse_jacobi_f('nc', 7, 1), 1))
        mpf('7.0000000000000009')
        sage: chop(ellipfun('nc', inverse_jacobi_f('nc', 7, 3), 3))
        mpf('7.0')
        sage: chop(ellipfun('nc', inverse_jacobi_f('nc', 0, 2), 2))
        mpf('0.0')
        sage: chop(ellipfun('nc', inverse_jacobi_f('nc', -18, -4), -4))
        mpf('-17.999999999999925')

        sage: # needs mpmath
        sage: chop(ellipfun('dn', inverse_jacobi_f('dn', -0.3, 1), 1))
        mpf('-0.29999999999999999')
        sage: chop(ellipfun('dn', inverse_jacobi_f('dn', 1, -1), -1))
        mpf('1.0')
        sage: chop(ellipfun('dn', inverse_jacobi_f('dn', 0.8, 0.5), 0.5))
        mpf('0.80000000000000004')
        sage: chop(ellipfun('dn', inverse_jacobi_f('dn', 5, -4), -4))
        mpf('5.0')
        sage: chop(ellipfun('dn', inverse_jacobi_f('dn', 0.4, 0.5), 0.5))
        mpf('0.40000000000000002')
        sage: chop(ellipfun('dn', inverse_jacobi_f('dn', -0.4, 0.5), 0.5))
        mpf('-0.40000000000000002')
        sage: chop(ellipfun('dn', inverse_jacobi_f('dn', -0.9, 0.5), 0.5))
        mpf('-0.90000000000000002')
        sage: chop(ellipfun('dn', inverse_jacobi_f('dn', -1.9, 0.2), 0.2))
        mpf('-1.8999999999999999')

        sage: # needs mpmath
        sage: chop(ellipfun('nd', inverse_jacobi_f('nd', -1.9, 1), 1))
        mpf('-1.8999999999999999')
        sage: chop(ellipfun('nd', inverse_jacobi_f('nd', 1, -1), -1))
        mpf('1.0')
        sage: chop(ellipfun('nd', inverse_jacobi_f('nd', 11, -6), -6))
        mpf('11.0')
        sage: chop(ellipfun('nd', inverse_jacobi_f('nd', 0, 8), 8))
        mpf('0.0')
        sage: chop(ellipfun('nd', inverse_jacobi_f('nd', -3, 0.8), 0.8))
        mpf('-2.9999999999999996')

        sage: # needs mpmath
        sage: chop(ellipfun('sc', inverse_jacobi_f('sc', -3, 0), 0))
        mpf('-3.0')
        sage: chop(ellipfun('sc', inverse_jacobi_f('sc', 2, 1), 1))
        mpf('2.0')
        sage: chop(ellipfun('sc', inverse_jacobi_f('sc', 0, 9), 9))
        mpf('0.0')
        sage: chop(ellipfun('sc', inverse_jacobi_f('sc', -7, 3), 3))
        mpf('-7.0')

        sage: # needs mpmath
        sage: chop(ellipfun('cs', inverse_jacobi_f('cs', -7, 0), 0))
        mpf('-6.9999999999999991')
        sage: chop(ellipfun('cs', inverse_jacobi_f('cs', 8, 1), 1))
        mpf('8.0')
        sage: chop(ellipfun('cs', inverse_jacobi_f('cs', 2, 6), 6))
        mpf('2.0')
        sage: chop(ellipfun('cs', inverse_jacobi_f('cs', 0, 4), 4))
        mpf('0.0')
        sage: chop(ellipfun('cs', inverse_jacobi_f('cs', -6, 8), 8))
        mpf('-6.0000000000000018')

        sage: chop(ellipfun('cd', inverse_jacobi_f('cd', -6, 0), 0))                    # needs mpmath
        mpf('-6.0000000000000009')
        sage: chop(ellipfun('cd', inverse_jacobi_f('cd', 1, 3), 3))                     # needs mpmath
        mpf('1.0')
        sage: chop(ellipfun('cd', inverse_jacobi_f('cd', 6, 8), 8))                     # needs mpmath
        mpf('6.0000000000000027')

        sage: chop(ellipfun('dc', inverse_jacobi_f('dc', 5, 0), 0))                     # needs mpmath
        mpf('5.0000000000000018')
        sage: chop(ellipfun('dc', inverse_jacobi_f('dc', -4, 2), 2))                    # needs mpmath
        mpf('-4.0000000000000018')

        sage: # needs mpmath
        sage: chop(ellipfun('sd', inverse_jacobi_f('sd', -4, 0), 0))
        mpf('-3.9999999999999991')
        sage: chop(ellipfun('sd', inverse_jacobi_f('sd', 7, 1), 1))
        mpf('7.0')
        sage: chop(ellipfun('sd', inverse_jacobi_f('sd', 0, 9), 9))
        mpf('0.0')
        sage: chop(ellipfun('sd', inverse_jacobi_f('sd', 8, 0.8), 0.8))
        mpf('7.9999999999999991')

        sage: chop(ellipfun('ds', inverse_jacobi_f('ds', 4, 0.25), 0.25))               # needs mpmath
        mpf('4.0')
    """
    from mpmath import mp as ctx
    prec = ctx.prec
    try:
        x = ctx.convert(x)
        m = ctx.convert(m)
        if not isinstance(x, ctx.mpf) or not isinstance(x, ctx.mpf):
            raise ValueError('arguments must be real')
        if kind == 'sn':
            if m == 0:
                return ctx.asin(x)
            elif m == 1:
                return ctx.atanh(x)
            elif x == 0:
                return ctx.zero
            sign = ctx.sign(x)  # sn is odd in x, so operate with abs(x) and
            x = abs(x)          # include the sign at the end
            if x <= 1:
                ctx.prec += 10
                phi = ctx.asin(x)
                return sign * ctx.ellipf(phi, m)
            elif x <= 1 / ctx.sqrt(m):
                K = ctx.ellipk(m)
                ctx.prec += 10
                xpn2 = x ** (-2)
                m1 = 1 - m
                ctx.prec += 10
                omxpn2 = 1 - xpn2
                ctx.prec += 10
                omxpn2dm1 = omxpn2 / m1
                ctx.prec += 10
                phi = ctx.asin(omxpn2dm1.sqrt())
                return sign * ctx.mpc(K, ctx.ellipf(phi, m1))
            else:
                ctx.prec += 10
                m1 = 1 - m
                K_prime = ctx.ellipk(m1)
                sqrtm = ctx.sqrt(m)
                ctx.prec += 10
                xsqrtm = x * sqrtm
                ctx.prec += 10
                phi = ctx.asin(1 / xsqrtm)
                ctx.prec += 10
                return sign * ctx.mpc(ctx.ellipf(phi, m), K_prime)
        if kind == 'ns':
            if m == 0:
                return ctx.acsc(x)
            elif m == 1:
                return ctx.acoth(x)
            elif x > 0:
                ctx.prec += 10
                return inverse_jacobi_f('sn', 1 / x, m)
            elif x == 0:
                ctx.prec += 10
                return ctx.j * ctx.ellipk(1 - m)
            else:
                ctx.prec += 10
                K_prime = ctx.ellipk(1 - m)
                odx = 1 / x
                ctx.prec += 10
                arcsnodx = inverse_jacobi_f('sn', odx, m)
                itK_prime = ctx.j * 2 * K_prime
                ctx.prec += 10
                return arcsnodx + itK_prime
        if kind == 'cn':
            if m == 0:
                return ctx.acos(x)
            elif m == 1:
                return ctx.asech(x)
            elif x == 1:
                return ctx.zero
            elif 0 <= x < 1:
                ctx.prec += 10
                x2 = x ** 2
                ctx.prec += 10
                osx2 = 1 - x2
                ctx.prec += 10
                return ctx.ellipf(ctx.asin(ctx.sqrt(osx2)), m)
            elif -1 <= x < 0:
                K = ctx.ellipk(m)
                ctx.prec += 10
                x2 = x ** 2
                ctx.prec += 10
                osx2 = 1 - x2
                ctx.prec += 10
                return (2 * K) - ctx.ellipf(ctx.asin(ctx.sqrt(osx2)), m)
            elif x > 1:
                ctx.prec += 10
                m1 = 1 - m
                xn2 = x ** (-2)
                ctx.prec += 10
                osx2 = 1 - xn2
                ctx.prec += 10
                return ctx.j * ctx.ellipf(ctx.asin(ctx.sqrt(osx2)), m1)
            elif x < -1:
                K = ctx.ellipk(m)
                ctx.prec += 10
                m1 = 1 - m
                xn2 = x ** (-2)
                tK = 2 * K
                ctx.prec += 10
                osx2 = 1 - xn2
                ctx.prec += 10
                phi = ctx.asin(ctx.sqrt(osx2))
                ctx.prec += 10
                return tK - ctx.j * ctx.ellipf(phi, m1)
        if kind == 'nc':
            if m == 0:
                return ctx.asec(x)
            elif m == 1:
                return ctx.acosh(x)
            elif x == 1:
                return ctx.zero
            elif x > 0:
                ctx.prec += 10
                return inverse_jacobi_f('cn', 1 / x, m)
            elif x == 0:
                ctx.prec += 10
                return ctx.j * ctx.ellipk(1 - m)
            else:
                K = ctx.ellipk(m)
                ctx.prec += 10
                K_prime = ctx.ellipk(1 - m)
                odx = 1 / x
                ctx.prec += 10
                arccnodx = inverse_jacobi_f('cn', odx, m)
                tK = 2 * K
                ctx.prec += 10
                return arccnodx - tK + ctx.j * 2 * K_prime
        if kind == 'dn':
            if x == 1:
                return ctx.zero
            if not m <= 1:
                raise ValueError('m must be <= 1')
            if m == 1:
                return ctx.asech(x)
            ctx.prec += 10
            m1 = 1 - m
            sqrtm1 = ctx.sqrt(m1)
            if sqrtm1 <= x < 1:
                ctx.prec += 10
                x2 = x ** 2
                ctx.prec += 10
                osx2 = 1 - x2
                ctx.prec += 10
                osx2dm = osx2 / m
                ctx.prec += 10
                return ctx.ellipf(ctx.asin(ctx.sqrt(osx2dm)), m)
            elif x > 1:
                ctx.prec += 10
                xn2 = x ** (-2)
                ctx.prec += 10
                osxn2 = 1 - xn2
                m1xn2 = m1 * xn2
                ctx.prec += 10
                osm1xn2 = 1 - m1xn2
                ctx.prec += 10
                sqrtosxn2dosm1xn2 = ctx.sqrt(osxn2 / osm1xn2)
                ctx.prec += 10
                return ctx.j * ctx.ellipf(ctx.asin(sqrtosxn2dosm1xn2), m1)
            elif 0 <= x < sqrtm1:
                K = ctx.ellipk(m)
                ctx.prec += 10
                x2 = x ** 2
                ctx.prec += 10
                x2dm1 = x2 / m1
                osx2 = 1 - x2
                ctx.prec += 10
                osx2dm1 = 1 - x2dm1
                ctx.prec += 10
                osx2dm1dosx2 = osx2dm1 / osx2
                ctx.prec += 10
                sqrtall = ctx.sqrt(osx2dm1dosx2)
                ctx.prec += 10
                phi = ctx.asin(sqrtall)
                ctx.prec += 10
                return K + ctx.j * ctx.ellipf(phi, m1)
            elif -sqrtm1 <= x < 0:
                K = ctx.ellipk(m)
                K_prime = ctx.ellipk(m1)
                ctx.prec += 10
                tK_prime = 2 * K_prime
                x2 = x ** 2
                ctx.prec += 10
                x2dm1 = x2 / m1
                osx2 = 1 - x2
                ctx.prec += 10
                osx2dm1 = 1 - x2dm1
                ctx.prec += 10
                osx2dm1dosx2 = osx2dm1 / osx2
                ctx.prec += 10
                sqrtall = ctx.sqrt(osx2dm1dosx2)
                ctx.prec += 10
                phi = ctx.asin(sqrtall)
                ctx.prec += 10
                return K + ctx.j * (tK_prime - ctx.ellipf(phi, m1))
            elif -1 <= x < -sqrtm1:
                K = ctx.ellipk(m)
                K_prime = ctx.ellipk(m1)
                ctx.prec += 10
                x2 = x ** 2
                tK = 2 * K
                # Note that the factor of 2 is missing in the reference
                # (formula (81)), probably mistakenly so
                tK_prime = 2 * K_prime
                ctx.prec += 10
                osx2 = 1 - x2
                ctx.prec += 10
                osx2dm = osx2 / m
                sqrtall = ctx.sqrt(osx2dm)
                ctx.prec += 10
                phi = ctx.asin(sqrtall)
                ctx.prec += 10
                return (tK - ctx.ellipf(phi, m)) + (ctx.j * tK_prime)
            elif x < -1:
                K = ctx.ellipk(m)
                K_prime = ctx.ellipk(m1)
                ctx.prec += 10
                tK = 2 * K
                tK_prime = 2 * K_prime
                xn2 = x ** (-2)
                ctx.prec += 10
                osxn2 = 1 - xn2
                m1xn2 = m1 * xn2
                ctx.prec += 10
                osm1xn2 = 1 - m1xn2
                ctx.prec += 10
                sqrtosxn2dosm1xn2 = ctx.sqrt(osxn2 / osm1xn2)
                ctx.prec += 10
                phi = ctx.asin(sqrtosxn2dosm1xn2)
                ctx.prec += 10
                return tK + ctx.j * (tK_prime - ctx.ellipf(phi, m1))
        if kind == 'nd':
            if m == 1:
                return ctx.acosh(x)
            elif x == 1:
                return ctx.zero
            elif x > 0:
                ctx.prec += 10
                return inverse_jacobi_f('dn', 1 / x, m)
            elif x == 0:
                ctx.prec += 10
                return ctx.j * ctx.ellipk(1 - m)
            else:
                K = ctx.ellipk(m)
                ctx.prec += 10
                tK = 2 * K
                ctx.prec += 10
                return inverse_jacobi_f('dn', 1 / x, m) - tK
        if kind == 'sc':
            if m == 0:
                return ctx.atan(x)
            elif m == 1:
                return ctx.asinh(x)
            elif x == 0:
                return ctx.zero
            else:
                ctx.prec += 10
                atanx = ctx.atan(x)
                return ctx.ellipf(atanx, m)
        if kind == 'cs':
            if m == 0:
                return ctx.acot(x)
            elif m == 1:
                return ctx.acsch(x)
            elif x > 0:
                ctx.prec += 10
                odx = 1 / x
                ctx.prec += 10
                return ctx.ellipf(ctx.atan(odx), m)
            elif x == 0:
                return ctx.ellipk(m)
            else:
                K = ctx.ellipk(m)
                ctx.prec += 10
                odx = 1 / x
                ctx.prec += 10
                phi = ctx.atan(odx)
                ctx.prec += 10
                return ctx.ellipf(phi, m) + (2 * K)
        if kind == 'cd':
            if m == 0:
                return ctx.acos(x)
            elif x == 1:
                return ctx.zero
            else:
                K = ctx.ellipk(m)
                ctx.prec += 10
                return inverse_jacobi_f('sn', x, m) - K
        if kind == 'dc':
            if m == 0:
                return ctx.asec(x)
            K = ctx.ellipk(m)
            ctx.prec += 10
            return inverse_jacobi_f('ns', x, m) - K
        if kind == 'sd':
            if m == 0:
                return ctx.asin(x)
            elif m == 1:
                return ctx.asinh(x)
            elif x == 0:
                return ctx.zero
            else:
                if m > 1:
                    raise ValueError('m must be <= 1')
                K = ctx.ellipk(m)
                ctx.prec += 10
                m1 = 1 - m
                ctx.prec += 10
                sqrtm1 = ctx.sqrt(m1)
                ctx.prec += 10
                xsqrtm1 = x * sqrtm1
                ctx.prec += 10
                return inverse_jacobi_f('cn', xsqrtm1, m) + K
        if kind == 'ds':
            if m == 0:
                return ctx.acsc(x)
            elif m == 1:
                return ctx.acsch(x)
            else:
                if m > 1:
                    raise ValueError('m must be <= 1')
                K = ctx.ellipk(m)
                ctx.prec += 10
                m1 = 1 - m
                ctx.prec += 10
                sqrtm1 = ctx.sqrt(m1)
                ctx.prec += 10
                xdsqrtm1 = x / sqrtm1
                ctx.prec += 10
                return inverse_jacobi_f('nc', xdsqrtm1, m) + K
    finally:
        ctx.prec = prec


def jacobi_am_f(x, m):
    r"""
    Internal function for numeric evaluation of the Jacobi amplitude function
    for real arguments. Procedure described in [Eh2013]_.

    TESTS::

        sage: # needs mpmath
        sage: from mpmath import ellipf
        sage: from sage.functions.jacobi import jacobi_am_f
        sage: ellipf(jacobi_am_f(0.5, 1), 1)
        mpf('0.5')
        sage: ellipf(jacobi_am(3, 0.3), 0.3)
        mpf('3.0')
        sage: ellipf(jacobi_am_f(2, -0.5), -0.5)
        mpf('2.0')
        sage: jacobi_am_f(2, -0.5)
        mpf('2.2680930777934176')
        sage: jacobi_am_f(-2, -0.5)
        mpf('-2.2680930777934176')
        sage: jacobi_am_f(-3, 2)
        mpf('0.36067407399586108')
    """
    from mpmath import mp as ctx
    prec = ctx.prec
    try:
        x = ctx.convert(x)
        m = ctx.convert(m)
        if not isinstance(x, ctx.mpf) or not isinstance(m, ctx.mpf):
            raise ValueError('arguments must be real')
        if abs(m) == 1:
            # gd(x)
            ctx.prec += 10
            tanhx = ctx.tanh(x)
            ctx.prec += 10
            return ctx.asin(tanhx)
        elif abs(m) > 1:
            ctx.prec += 10
            # Real values needed for atan2; as per "Handbook of Elliptic
            # Integrals for Engineers and Scientists" 121.02, sn is real for
            # real x. The imaginary components can thus be safely discarded.
            snx = ctx.ellipfun('sn', x, m).real
            cnx = ctx.ellipfun('cn', x, m).real
            ctx.prec += 10
            return ctx.atan2(snx, cnx)
        else:
            ctx.prec += 10
            K = ctx.ellipk(m)
            if abs(x) <= K:
                snx = ctx.ellipfun('sn', x, m).real
                cnx = ctx.ellipfun('cn', x, m).real
                ctx.prec += 10
                return ctx.atan2(snx, cnx)
            else:
                # Do argument reduction on x to end up with z = x - 2nK, with
                # abs(z) <= K
                ctx.prec += 10
                tK = 2 * K
                ctx.prec += 10
                n = ctx.floor(x / tK)
                ctx.prec += 10
                tnK = n * tK
                npi = n * ctx.pi()
                ctx.prec += 10
                z = x - tnK
                ctx.prec += 10
                # z (and therefore sn(z, m) and cn(z, m)) is real because K(m)
                # is real for abs(m) <= 1.
                snz = ctx.ellipfun('sn', z, m).real
                cnz = ctx.ellipfun('cn', z, m).real
                ctx.prec += 10
                return ctx.atan2(snz, cnz) + npi
    finally:
        ctx.prec = prec



# sage_setup: distribution = sagemath-categories
"""
Tests for rings

TESTS::

    sage: K.<x> = FractionField(QQ['x'])
    sage: V.<z> = K[]
    sage: x+z
    z + x
"""

import sage.misc.prandom as random

from sage.misc.random_testing import random_testing


def prime_finite_field():
    """
    Create a random prime finite field with cardinality at most 10^20.

    OUTPUT: a prime finite field

    EXAMPLES::

        sage: import sage.rings.tests
        sage: K = sage.rings.tests.prime_finite_field(); K                              # needs sage.rings.finite_rings
        Finite Field of size ...
        sage: K.cardinality().is_prime()                                                # needs sage.rings.finite_rings
        True
    """
    from sage.rings.integer_ring import ZZ
    from sage.rings.finite_rings.finite_field_constructor import GF
    return GF(ZZ.random_element(x=2, y=10**20 - 12).next_prime())


def finite_field():
    """
    Create a random finite field with degree at most 20 and prime at most 10^6.

    OUTPUT: a finite field

    EXAMPLES::

        sage: import sage.rings.tests
        sage: K = sage.rings.tests.finite_field(); K                                    # needs sage.rings.finite_rings
        Finite Field...of size ...
        sage: K.cardinality().is_prime_power()                                          # needs sage.rings.finite_rings
        True
        sage: while K.cardinality().is_prime():                                         # needs sage.rings.finite_rings
        ....:     K = sage.rings.tests.finite_field()
    """
    from sage.rings.integer_ring import ZZ
    from sage.rings.finite_rings.finite_field_constructor import GF
    p = ZZ.random_element(x=2, y=10**6 - 18).next_prime()
    d = ZZ.random_element(x=1, y=20)
    return GF(p**d, 'a')


def small_finite_field():
    """
    Create a random finite field with cardinality at most 2^16.

    OUTPUT: a finite field

    EXAMPLES::

        sage: # needs sage.rings.finite_rings
        sage: import sage.rings.tests
        sage: K = sage.rings.tests.small_finite_field(); K
        Finite Field...of size ...
        sage: q = K.cardinality()
        sage: q.is_prime_power()
        True
        sage: q <= 2^16
        True
    """
    from sage.rings.integer_ring import ZZ
    from sage.rings.finite_rings.finite_field_constructor import GF
    while True:
        q = ZZ.random_element(x=2, y=2**16)
        if q.is_prime_power():
            return GF(q, 'a')


def integer_mod_ring():
    """
    Return a random ring of integers modulo n with n at most 50000.

    EXAMPLES::

        sage: import sage.rings.tests
        sage: R = sage.rings.tests.integer_mod_ring(); R
        Ring of integers modulo ...
        sage: R.cardinality() <= 50000
        True
    """
    from sage.rings.integer_ring import ZZ
    from sage.rings.finite_rings.integer_mod_ring import IntegerModRing
    n = ZZ.random_element(x=2, y=50000)
    return IntegerModRing(n)


def padic_field():
    """
    Return a random `p`-adic field modulo n with p at most 10000
    and precision between 10 and 100.

    EXAMPLES::

        sage: import sage.rings.tests
        sage: sage.rings.tests.padic_field()                                            # needs sage.rings.padics
        ...-adic Field with capped relative precision ...
    """
    from sage.rings.integer_ring import ZZ
    from sage.rings.padics.factory import Qp
    prec = ZZ.random_element(x=10, y=100)
    p = ZZ.random_element(x=2, y=10**4 - 30).next_prime()
    return Qp(p, prec)


def quadratic_number_field():
    """
    Return a quadratic extension of QQ.

    EXAMPLES::

        sage: import sage.rings.tests
        sage: K = sage.rings.tests.quadratic_number_field(); K                          # needs sage.rings.number_field
        Number Field in a with defining polynomial x^2 ... with a = ...
    """
    from sage.rings.integer_ring import ZZ
    from sage.rings.number_field.number_field import QuadraticField
    while True:
        d = ZZ.random_element(x=-10**5, y=10**5)
        if not d.is_square():
            return QuadraticField(d, 'a')


def absolute_number_field(maxdeg=10):
    """
    Return an absolute extension of QQ of degree at most 10.

    EXAMPLES::

        sage: import sage.rings.tests
        sage: K = sage.rings.tests.absolute_number_field(); K                           # needs sage.rings.number_field
        Number Field in a with defining polynomial ...
        sage: K.degree() <= 10                                                          # needs sage.rings.number_field
        True
    """
    from sage.rings.integer_ring import ZZ
    from sage.rings.number_field.number_field import NumberField
    R = ZZ['x']
    while True:
        f = R.random_element(degree=ZZ.random_element(x=1, y=maxdeg),
                             x=-100, y=100)
        if f.degree() <= 0:
            continue
        f = f + R.gen()**(f.degree() + 1)  # make monic
        if f.is_irreducible():
            return NumberField(f, 'a')


def relative_number_field(n=2, maxdeg=2):
    """
    Return a tower of at most n extensions each of degree at most maxdeg.

    EXAMPLES::

        sage: # needs sage.rings.number_field
        sage: import sage.rings.tests
        sage: K = sage.rings.tests.relative_number_field(3); K
        Number Field in aaa with defining polynomial x^2 ... over its base field
        sage: K.relative_degree()
        2
        sage: L = K.base_ring()
        sage: L.relative_degree()
        2
        sage: M = L.base_ring()
        sage: M.relative_degree()
        2
        sage: M.base_ring() is QQ
        True

    TESTS:

    Check that :issue:`32117` is fixed::

        sage: set_random_seed(3030)
        sage: from sage.rings.tests import relative_number_field
        sage: _ = relative_number_field(3)                                              # needs sage.rings.number_field
    """
    from sage.rings.integer_ring import ZZ
    K = absolute_number_field(maxdeg)
    n -= 1
    var = 'aa'
    R = ZZ['x']
    R1 = K['x']
    while n >= 1:
        while True:
            f = R.random_element(degree=ZZ.random_element(x=1, y=maxdeg),
                                 x=-100, y=100)
            if f.degree() <= 0:
                continue
            f = f * f.denominator()  # bug trac #4781
            f = f + R.gen()**maxdeg  # make monic
            if R1(f).is_irreducible():
                break
        K = K.extension(f, var)
        R1 = K['x']
        var += 'a'
        n -= 1
    return K


def rings0():
    """
    Return a list of pairs (f, desc), where f is a function that when
    called creates a random ring of a certain representative type
    described by desc.

    RINGS:

    - ZZ
    - QQ
    - ZZ/nZZ
    - GF(p)
    - GF(q)
    - p-adic fields
    - quadratic number fields
    - absolute number fields
    - relative number fields

    EXAMPLES::

        sage: import sage.rings.tests
        sage: type(sage.rings.tests.rings0())
        <... 'list'>
    """
    from sage.rings.integer_ring import IntegerRing
    from sage.rings.rational_field import RationalField

    v = [(IntegerRing, 'ring of integers'),
         (RationalField, 'field of rational numbers'),
         (integer_mod_ring, 'integers modulo n for n at most 50000')]
    try:
        v += [(prime_finite_field, 'a prime finite field with cardinality at most 10^20'),
              (finite_field, 'finite field with degree at most 20 and prime at most 10^6'),
              (small_finite_field, 'finite field with cardinality at most 2^16')]
    except ImportError:
        pass

    try:
        v += [(padic_field, 'a p-adic field')]
    except ImportError:
        pass

    try:
        v += [(quadratic_number_field, 'a quadratic number field'),
              (absolute_number_field, 'an absolute number field of degree at most 10'),
              (relative_number_field, 'a tower of at most 2 extensions each of degree at most 2')]
    except ImportError:
        pass

    return v


def rings1():
    """
    Return an iterator over random rings.

    Return a list of pairs (f, desc), where f is a function that
    outputs a random ring that takes a ring and possibly
    some other data as constructor.

    RINGS:

    - polynomial ring in one variable over a rings0() ring.
    - polynomial ring over a rings1() ring.
    - multivariate polynomials
    - power series rings in one variable over a rings0() ring.

    EXAMPLES::

        sage: import sage.rings.tests
        sage: type(sage.rings.tests.rings0())
        <... 'list'>
    """
    v = rings0()
    X = random_rings(level=0)
    from sage.rings.polynomial.polynomial_ring_constructor import PolynomialRing
    from sage.rings.power_series_ring import PowerSeriesRing
    from sage.rings.integer_ring import ZZ

    v = [(lambda: PolynomialRing(next(X), names='x'),
          'univariate polynomial ring over level 0 ring'),
         (lambda: PowerSeriesRing(next(X), names='x'),
          'univariate power series ring over level 0 ring')]

    try:
        from sage.rings.polynomial.laurent_polynomial_ring import LaurentPolynomialRing
    except ImportError:
        pass
    else:
        v += [(lambda: LaurentPolynomialRing(next(X), names='x'),
               'univariate Laurent polynomial ring over level 0 ring')]

    v += [(lambda: PolynomialRing(next(X), abs(ZZ.random_element(x=2, y=10)),
                                  names='x'),
           'multivariate polynomial ring in between 2 and 10 variables over a level 0 ring')]

    return v


MAX_LEVEL = 99999


def random_rings(level=MAX_LEVEL):
    """
    Return an iterator over random rings up to the given "level" of complexity.

    EXAMPLES::

        sage: import sage.rings.tests
        sage: type(sage.rings.tests.random_rings())
        <... 'generator'>
    """
    v = rings0()
    if level >= 1:
        v += rings1()
    while True:
        yield random.choice(v)[0]()


@random_testing
def test_random_elements(level=MAX_LEVEL, trials=1):
    """
    Create random elements of random rings until a crash occurs, in
    which case an exception is raised.  Defaults to running a single
    trial, but more can be specified.  To run tests in an infinite
    loop, you could use::

        while True: test_random_elements(trials=100, print_seed=True)

    INPUT:

    - ``level`` -- (default: ``MAX_LEVEL``) controls the types of rings to use
    - ``trials`` -- a positive integer (default: 1); the number of trials to run
    - ``seed`` -- the random seed to use; if not specified, uses a truly random seed
    - ``print_seed`` -- if ``True`` (default: ``False``), prints the random seed chosen

    EXAMPLES::

        sage: import sage.rings.tests
        sage: sage.rings.tests.test_random_elements(trials=2, seed=0)                   # needs sage.rings.number_field
        survived 0 tests
        Rational Field
        -1/2
        ----
        survived 1 tests
        Number Field in a with defining polynomial x^2 - 61891 with a = 248.7790184079036?
        -12
        ----

        sage: # needs sage.rings.finite_rings sage.rings.number_field sage.rings.padics
        sage: sage.rings.tests.test_random_elements(trials=10)
        survived 0 tests...
        sage: sage.rings.tests.test_random_elements(trials=1000)  # long time (5 seconds)
        survived 0 tests...
    """
    r = random_rings(level)
    i = 0
    for R in r:
        print("survived %s tests" % i)
        i += 1
        print(R)
        print(R.random_element())
        print("----")
        if i >= trials:
            return


@random_testing
def test_random_arith(level=MAX_LEVEL, trials=1):
    """
    Create random elements of random rings and do some arithmetic with them.

    Repeats until a crash occurs, in which case an exception is
    raised.  Defaults to running a single trial, but more can be
    specified.  To run tests in an infinite loop, you could use::

        while True: test_random_arith(trials=100, print_seed=True)

    INPUT:

    - ``level`` -- (default: ``MAX_LEVEL``) controls the types of rings to use
    - ``trials`` -- positive integer (default: 1); the number of trials to run
    - ``seed`` -- the random seed to use; if not specified, uses a truly random seed
    - ``print_seed`` -- if ``True`` (default: ``False``), prints the random seed chosen

    EXAMPLES::

        sage: # needs sage.rings.finite_rings sage.rings.number_field sage.rings.padics
        sage: import sage.rings.tests
        sage: sage.rings.tests.test_random_arith(trials=2, seed=0)
        survived 0 tests
        Rational Field
        -1/2 -1/95
        49/95
        survived 1 tests
        Number Field in a with defining polynomial x^2 - 15083 with a = 122.81286577553673?
        a -2*a - 1
        2*a - 30164
        sage: sage.rings.tests.test_random_arith(trials=10)
        survived 0 tests...
        sage: sage.rings.tests.test_random_arith(trials=1000)   # long time (5 seconds?)
        survived 0 tests...
    """
    i = 0
    for x in random_rings(level):
        print("survived %s tests" % i)
        i += 1
        print(x)
        a = x.random_element()
        b = x.random_element()
        print(a, b)
        print(a * b + a - b + 1)
        if i >= trials:
            return


@random_testing
def test_karatsuba_multiplication(base_ring, maxdeg1, maxdeg2,
                                  ref_mul=lambda f, g: f._mul_generic(g),
                                  base_ring_random_elt_args=[],
                                  numtests=10, verbose=False):
    """
    Test univariate Karatsuba multiplication against other multiplication algorithms.

    EXAMPLES:

    First check that random tests are reproducible::

        sage: from sage.rings.tests import test_karatsuba_multiplication
        sage: test_karatsuba_multiplication(ZZ, 6, 5, verbose=True, seed=42)
        test_karatsuba_multiplication: ring=Univariate Polynomial Ring in x over Integer Ring, threshold=2
          (x^6 + 4*x^5 + 4*x^4 - 3*x^3 - x^2 - x)*(2*x^4 + 3*x^3 - 20*x^2 - 2*x + 1)
          (4*x^5 + 16*x^2 + x - 41)*(x^2 + x - 1)
          (8*x^2 + 2*x + 1)*(3)
          (-4*x - 1)*(-8*x^2 - x)
          (-x^6 - x^3 - x^2 + x + 1)*(2*x^3 - x + 3)
          (-x^2 + x + 1)*(x^4 + x^3 - x^2 - x + 76)
          (4*x^3 + x^2 + 6)*(-x^2 - 5*x)
          (x + 4)*(-x + 5)
          (-2*x)*(3*x^2 - x)
          (x^6 + 21*x^5 + x^4 + 4*x^3 - x^2)*(14*x^4 + x^3 + 2*x^2 - 12*x)

    Test Karatsuba multiplication of polynomials of small degree over some common rings::

        sage: rings = [QQ]
        sage: rings += [ZZ[I], ZZ[I, sqrt(2)]]                                          # needs sage.rings.number_field sage.symbolic
        sage: rings += [GF(49, 'a')]                                                    # needs sage.rings.finite_rings
        sage: rings += [MatrixSpace(GF(17), 3)]                                         # needs sage.modules
        sage: for C in rings:                                                           # needs sage.modules
        ....:     test_karatsuba_multiplication(C, 10, 10)

    Zero-tests over ``QQbar`` are currently very slow, so we test only very small examples::

        sage: test_karatsuba_multiplication(QQbar, 3, 3, numtests=2)    # long time, needs sage.rings.number_field

    Larger degrees (over ``ZZ``, using FLINT)::

        sage: test_karatsuba_multiplication(ZZ, 1000, 1000,
        ....:                               ref_mul=lambda f,g: f*g,
        ....:                               base_ring_random_elt_args=[1000])

    Some more aggressive tests::

        sage: testrings = [ZZ[I, sqrt(2)], ZZ[I, sqrt(2), sqrt(3)]]     # long time
        sage: for C in testrings:                                       # long time
        ....:     test_karatsuba_multiplication(C, 100, 100)
        sage: test_karatsuba_multiplication(ZZ, 10000, 10000,           # long time
        ....:                               ref_mul=lambda f,g: f*g,
        ....:                               base_ring_random_elt_args=[100000])
    """
    from sage.misc.prandom import randint
    from sage.misc.sage_input import sage_input
    from sage.rings.polynomial.polynomial_ring_constructor import PolynomialRing
    threshold = randint(0, min(maxdeg1, maxdeg2))
    R = PolynomialRing(base_ring, 'x')
    if verbose:
        print(f"test_karatsuba_multiplication: ring={R}, threshold={threshold}")
    for _ in range(numtests):
        f = R.random_element(randint(0, maxdeg1), False, *base_ring_random_elt_args)
        g = R.random_element(randint(0, maxdeg2), False, *base_ring_random_elt_args)
        if verbose:
            print("  ({})*({})".format(f, g))
        if ref_mul(f, g) - f._mul_karatsuba(g, threshold) != 0:
            msg = "Multiplication failed for elements defined by\n"
            msg += f"{sage_input(f)}\n"
            msg += "and\n"
            msg += f"{sage_input(g)}"
            raise ValueError(msg)
    return



# sage_setup: distribution = sagemath-categories
r"""
Partition backtrack functions for sets

EXAMPLES::

    sage: import sage.groups.perm_gps.partn_ref.refinement_sets

REFERENCE:

    [1] McKay, Brendan D. Practical Graph Isomorphism. Congressus Numerantium,
        Vol. 30 (1981), pp. 45-87.

    [2] Leon, Jeffrey. Permutation Group Algorithms Based on Partitions, I:
        Theory and Algorithms. J. Symbolic Computation, Vol. 12 (1991), pp.
        533-583.
"""

#*****************************************************************************
#       Copyright (C) 2006 - 2011 Robert L. Miller <rlmillster@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  http://www.gnu.org/licenses/
#*****************************************************************************

from sage.groups.perm_gps.partn_ref.data_structures cimport *
from sage.groups.perm_gps.partn_ref.double_coset cimport double_coset
from sage.data_structures.bitset_base cimport *


def set_stab_py(generators, sett, relab=False):
    r"""
    Compute the setwise stabilizer of a subset of [0..n-1] in a subgroup of S_n.

    EXAMPLES::

        sage: from sage.groups.perm_gps.partn_ref.refinement_sets import set_stab_py

    Degree 4 examples

    A four-cycle::

        sage: set_stab_py([[1,2,3,0]], [0])
        []
        sage: set_stab_py([[1,2,3,0]], [0,1])
        []
        sage: set_stab_py([[1,2,3,0]], [0,1,2])
        []
        sage: set_stab_py([[1,2,3,0]], [0,1,2,3])
        [[1, 2, 3, 0]]
        sage: set_stab_py([[1,2,3,0]], [0,2])
        [[2, 3, 0, 1]]

    Symmetric group::

        sage: set_stab_py([[1,0,2,3],[1,2,3,0]], [0])
        [[0, 1, 3, 2], [0, 2, 1, 3]]
        sage: set_stab_py([[1,0,2,3],[1,2,3,0]], [0,1])
        [[1, 0, 2, 3], [0, 1, 3, 2]]
        sage: set_stab_py([[1,0,2,3],[1,2,3,0]], [0,1,2,3])
        [[0, 1, 3, 2], [0, 2, 1, 3], [1, 0, 2, 3]]
        sage: set_stab_py([[1,0,2,3],[1,2,3,0]], [0,3])
        [[3, 1, 2, 0], [0, 2, 1, 3]]

    Klein 4-group::

        sage: set_stab_py([[1,0,2,3],[0,1,3,2]], [0])
        [[0, 1, 3, 2]]
        sage: set_stab_py([[1,0,2,3],[0,1,3,2]], [0,1])
        [[0, 1, 3, 2], [1, 0, 2, 3]]
        sage: set_stab_py([[1,0,2,3],[0,1,3,2]], [0,2])
        []

    Dihedral group::

        sage: set_stab_py([[1,2,3,0],[0,3,2,1]], [0])
        [[0, 3, 2, 1]]
        sage: set_stab_py([[1,2,3,0],[0,3,2,1]], [0,1])
        [[1, 0, 3, 2]]
        sage: set_stab_py([[1,2,3,0],[0,3,2,1]], [0,2])
        [[2, 1, 0, 3], [0, 3, 2, 1]]
        sage: set_stab_py([[1,2,3,0],[0,3,2,1]], [1])
        [[2, 1, 0, 3]]
        sage: set_stab_py([[1,2,3,0],[0,3,2,1]], [1,2,3])
        [[0, 3, 2, 1]]

    Alternating group::

        sage: set_stab_py([[1,2,0,3],[0,2,3,1]], [0])
        [[0, 2, 3, 1]]
        sage: set_stab_py([[1,2,0,3],[0,2,3,1]], [1])
        [[2, 1, 3, 0]]
        sage: set_stab_py([[1,2,0,3],[0,2,3,1]], [2])
        [[1, 3, 2, 0]]
        sage: set_stab_py([[1,2,0,3],[0,2,3,1]], [3])
        [[1, 2, 0, 3]]
        sage: set_stab_py([[1,2,0,3],[0,2,3,1]], [0,1])
        [[1, 0, 3, 2]]
        sage: set_stab_py([[1,2,0,3],[0,2,3,1]], [0,2])
        [[2, 3, 0, 1]]
        sage: set_stab_py([[1,2,0,3],[0,2,3,1]], [0,3])
        [[3, 2, 1, 0]]

    Larger degree examples

    Dihedral group of degree 5::

        sage: set_stab_py([[1,2,3,4,0],[0,4,3,2,1]], [])
        [[0, 4, 3, 2, 1], [1, 0, 4, 3, 2]]
        sage: set_stab_py([[1,2,3,4,0],[0,4,3,2,1]], [0])
        [[0, 4, 3, 2, 1]]
        sage: set_stab_py([[1,2,3,4,0],[0,4,3,2,1]], [0,2])
        [[2, 1, 0, 4, 3]]

    Dihedral group of degree 6::

        sage: set_stab_py([[1,2,3,4,5,0],[0,5,4,3,2,1]], [])
        [[0, 5, 4, 3, 2, 1], [1, 0, 5, 4, 3, 2]]
        sage: set_stab_py([[1,2,3,4,5,0],[0,5,4,3,2,1]], [0])
        [[0, 5, 4, 3, 2, 1]]
        sage: set_stab_py([[1,2,3,4,5,0],[0,5,4,3,2,1]], [0,1])
        [[1, 0, 5, 4, 3, 2]]
        sage: set_stab_py([[1,2,3,4,5,0],[0,5,4,3,2,1]], [0,2])
        [[2, 1, 0, 5, 4, 3]]
        sage: set_stab_py([[1,2,3,4,5,0],[0,5,4,3,2,1]], [0,3])
        [[0, 5, 4, 3, 2, 1], [3, 2, 1, 0, 5, 4]]
        sage: set_stab_py([[1,2,3,4,5,0],[0,5,4,3,2,1]], [0,2,4])
        [[2, 1, 0, 5, 4, 3], [4, 3, 2, 1, 0, 5]]

    Canonical labels::

        sage: set_stab_py([[0,2,1,4,3,5,8,7,6],[8,7,6,3,5,4,2,1,0]], [0,1,3,5,6], True)
        ([], [7, 8, 6, 3, 4, 5, 2, 0, 1])
        sage: set_stab_py([[0,2,1,4,3,5,8,7,6],[8,7,6,3,5,4,2,1,0]], [0,3,5,6,8], True)
        ([], [2, 1, 0, 5, 4, 3, 7, 6, 8])
    """
    if len(generators) == 0:
        return []
    cdef int i, j, n = len(generators[0]), n_gens = len(generators)
    cdef StabilizerChain *supergroup = SC_new(n)
    cdef aut_gp_and_can_lab *stabilizer
    cdef int *gens = <int *> sig_malloc(n*n_gens * sizeof(int))
    cdef subset *subset_sett = <subset *> sig_malloc(sizeof(subset))
    if gens is NULL or supergroup is NULL or subset_sett is NULL:
        SC_dealloc(supergroup)
        sig_free(gens)
        sig_free(subset_sett)
        raise MemoryError
    bitset_init(&subset_sett.bits, n)
    subset_sett.scratch = <int *> sig_malloc((3*n+1) * sizeof(int))
    for i from 0 <= i < len(generators):
        for j from 0 <= j < n:
            gens[n*i + j] = generators[i][j]
    if SC_insert(supergroup, 0, gens, n_gens):
        SC_dealloc(supergroup)
        sig_free(gens)
        sig_free(subset_sett)
        raise MemoryError
    sig_free(gens)
    bitset_clear(&subset_sett.bits)
    for i in sett:
        bitset_add(&subset_sett.bits, i)
    stabilizer = set_stab(supergroup, subset_sett, relab)
    SC_dealloc(supergroup)
    bitset_free(&subset_sett.bits)
    sig_free(subset_sett.scratch)
    sig_free(subset_sett)
    if stabilizer is NULL:
        raise MemoryError
    stab_gens = []
    for i from 0 <= i < stabilizer.num_gens:
        stab_gens.append([stabilizer.generators[i*n+j] for j from 0 <= j < n])
    if relab:
        relabeling = [stabilizer.relabeling[j] for j from 0 <= j < n]
    deallocate_agcl_output(stabilizer)
    if relab:
        return stab_gens, relabeling
    return stab_gens


cdef aut_gp_and_can_lab *set_stab(StabilizerChain *supergroup, subset *sett, bint relab) noexcept:
    r"""
    Compute the set stabilizer of ``sett`` within ``supergroup``. (Note that
    ``set`` is a reserved Python keyword.) If ``relab`` is specified then
    compute the canonical label of the set under the action of the group.
    """
    cdef aut_gp_and_can_lab *output
    cdef int n = supergroup.degree
    cdef PartitionStack *part = PS_new(n, 1)
    if part is NULL:
        return NULL
    output = get_aut_gp_and_can_lab(<void *> sett, part, n,
        &all_set_children_are_equivalent, &refine_set, &compare_sets, relab,
        supergroup, NULL, NULL)
    PS_dealloc(part)
    if output is NULL:
        return NULL
    return output


def sets_isom_py(generators, set1, set2):
    r"""
    Compute whether ``set1`` and ``set2`` are isomorphic under the action of
    the group generated by the generators given in list form.

    EXAMPLES::

        sage: from sage.groups.perm_gps.partn_ref.refinement_sets import sets_isom_py

    Degree 3 examples

    Trivial group::

        sage: sets_isom_py([], [0,1,2], [0,1])
        False
        sage: sets_isom_py([], [0,1,2], [0,2,1])
        [0, 1, 2]
        sage: sets_isom_py([[0,1,2]], [0,1,2], [0,2,1])
        [0, 1, 2]
        sage: sets_isom_py([[0,1,2]], [0], [0])
        [0, 1, 2]
        sage: sets_isom_py([[0,1,2]], [0], [1])
        False
        sage: sets_isom_py([[0,1,2]], [0], [2])
        False
        sage: sets_isom_py([[0,1,2]], [0,1], [1,0])
        [0, 1, 2]

    Three-cycle::

        sage: sets_isom_py([[1,2,0]], [0], [1])
        [1, 2, 0]
        sage: sets_isom_py([[1,2,0]], [0], [2])
        [2, 0, 1]
        sage: sets_isom_py([[1,2,0]], [0], [0])
        [0, 1, 2]
        sage: sets_isom_py([[1,2,0]], [0,1], [0])
        False
        sage: sets_isom_py([[1,2,0]], [0,1], [1])
        False
        sage: sets_isom_py([[1,2,0]], [0,1], [2])
        False
        sage: sets_isom_py([[1,2,0]], [0,1], [0,2])
        [2, 0, 1]
        sage: sets_isom_py([[1,2,0]], [0,1], [1,2])
        [1, 2, 0]
        sage: sets_isom_py([[1,2,0]], [0,1], [1,0])
        [0, 1, 2]
        sage: sets_isom_py([[1,2,0]], [0,2,1], [2,1,0])
        [0, 1, 2]

    Transposition::

        sage: sets_isom_py([[1,0,2]], [0], [])
        False
        sage: sets_isom_py([[1,0,2]], [0], [1,2])
        False
        sage: sets_isom_py([[1,0,2]], [0], [1])
        [1, 0, 2]
        sage: sets_isom_py([[1,0,2]], [0], [0])
        [0, 1, 2]
        sage: sets_isom_py([[1,0,2]], [0,1], [2])
        False
        sage: sets_isom_py([[1,0,2]], [0,2], [1,2])
        [1, 0, 2]
        sage: sets_isom_py([[1,0,2]], [0], [2])
        False
        sage: sets_isom_py([[1,0,2]], [0,1], [1,2])
        False

    Symmetric group S_3::

        sage: sets_isom_py([[1,0,2],[1,2,0]], [], [])
        [0, 1, 2]
        sage: sets_isom_py([[1,0,2],[1,2,0]], [0], [])
        False
        sage: sets_isom_py([[1,0,2],[1,2,0]], [0], [0])
        [0, 1, 2]
        sage: sets_isom_py([[1,0,2],[1,2,0]], [0], [1])
        [1, 0, 2]
        sage: sets_isom_py([[1,0,2],[1,2,0]], [0], [2])
        [2, 0, 1]
        sage: sets_isom_py([[1,0,2],[1,2,0]], [0,2], [2])
        False
        sage: sets_isom_py([[1,0,2],[1,2,0]], [0,2], [1,2])
        [1, 0, 2]
        sage: sets_isom_py([[1,0,2],[1,2,0]], [0,2], [0,1])
        [0, 2, 1]
        sage: sets_isom_py([[1,0,2],[1,2,0]], [0,2], [0,2])
        [0, 1, 2]
        sage: sets_isom_py([[1,0,2],[1,2,0]], [0,2], [0,1,2])
        False
        sage: sets_isom_py([[1,0,2],[1,2,0]], [0,2,1], [0,1,2])
        [0, 1, 2]

    Degree 4 examples

    Trivial group::

        sage: sets_isom_py([[0,1,2,3]], [], [])
        [0, 1, 2, 3]
        sage: sets_isom_py([[0,1,2,3]], [0], [])
        False
        sage: sets_isom_py([[0,1,2,3]], [0], [1])
        False
        sage: sets_isom_py([[0,1,2,3]], [0], [2])
        False
        sage: sets_isom_py([[0,1,2,3]], [0], [3])
        False
        sage: sets_isom_py([[0,1,2,3]], [0], [0])
        [0, 1, 2, 3]
        sage: sets_isom_py([[0,1,2,3]], [0,1], [1,2])
        False
        sage: sets_isom_py([[0,1,2,3]], [0,1], [0,1])
        [0, 1, 2, 3]
        sage: sets_isom_py([[0,1,2,3]], [0,1,2,3], [0,1])
        False
        sage: sets_isom_py([[0,1,2,3]], [0,1,2,3], [0,1,2,3])
        [0, 1, 2, 3]

    Four-cycle::

        sage: sets_isom_py([[1,2,3,0]], [0,1,2,3], [0,1,2,3])
        [0, 1, 2, 3]
        sage: sets_isom_py([[1,2,3,0]], [], [])
        [0, 1, 2, 3]
        sage: sets_isom_py([[1,2,3,0]], [0], [0])
        [0, 1, 2, 3]
        sage: sets_isom_py([[1,2,3,0]], [0], [1])
        [1, 2, 3, 0]
        sage: sets_isom_py([[1,2,3,0]], [0], [2])
        [2, 3, 0, 1]
        sage: sets_isom_py([[1,2,3,0]], [0], [3])
        [3, 0, 1, 2]
        sage: sets_isom_py([[1,2,3,0]], [0,1], [3])
        False
        sage: sets_isom_py([[1,2,3,0]], [0,1], [])
        False
        sage: sets_isom_py([[1,2,3,0]], [0,1], [1,2,3])
        False
        sage: sets_isom_py([[1,2,3,0]], [0,1], [1,2])
        [1, 2, 3, 0]
        sage: sets_isom_py([[1,2,3,0]], [0,1], [2,3])
        [2, 3, 0, 1]
        sage: sets_isom_py([[1,2,3,0]], [0,1], [0,3])
        [3, 0, 1, 2]
        sage: sets_isom_py([[1,2,3,0]], [0,2], [0,2])
        [0, 1, 2, 3]
        sage: sets_isom_py([[1,2,3,0]], [0,2], [1,3])
        [3, 0, 1, 2]
        sage: sets_isom_py([[1,2,3,0]], [0,1,2], [1,2,3])
        [1, 2, 3, 0]
        sage: sets_isom_py([[1,2,3,0]], [0,1,2], [0,2,3])
        [2, 3, 0, 1]
        sage: sets_isom_py([[1,2,3,0]], [0,1,2], [0,1,3])
        [3, 0, 1, 2]
        sage: sets_isom_py([[1,2,3,0]], [0,1,2], [0,1,2])
        [0, 1, 2, 3]

    Two transpositions::

        sage: from sage.groups.perm_gps.partn_ref.refinement_sets import sets_isom_py
        sage: sets_isom_py([[2,3,0,1]], [0], [2])
        [2, 3, 0, 1]
        sage: sets_isom_py([[2,3,0,1]], [1], [3])
        [2, 3, 0, 1]
        sage: sets_isom_py([[2,3,0,1]], [1], [1])
        [0, 1, 2, 3]
        sage: sets_isom_py([[2,3,0,1]], [1], [2])
        False
        sage: sets_isom_py([[2,3,0,1]], [0,3], [1,2])
        [2, 3, 0, 1]
        sage: sets_isom_py([[2,3,0,1]], [0,3], [3,0])
        [0, 1, 2, 3]
        sage: sets_isom_py([[2,3,0,1]], [0,1,3], [0,2,3])
        False
        sage: sets_isom_py([[2,3,0,1]], [0,1,3], [1,2,3])
        [2, 3, 0, 1]
    """
    set1 = set(set1)
    set2 = set(set2)
    if not generators:
        if set1 == set2:
            return list(range(max(set1) + 1))
        else:
            return False

    cdef int i, j, n = len(generators[0]), n_gens = len(generators)
    cdef StabilizerChain *supergroup = SC_new(n)
    cdef int *gens = <int *> sig_malloc(n*n_gens * sizeof(int))
    cdef int *isom = <int *> sig_malloc(n * sizeof(int))
    cdef subset *subset_sett1 = <subset *> sig_malloc(sizeof(subset))
    cdef subset *subset_sett2 = <subset *> sig_malloc(sizeof(subset))
    bitset_init(&subset_sett1.bits, n)
    bitset_init(&subset_sett2.bits, n)
    subset_sett1.scratch = <int *> sig_malloc((3*n+1) * sizeof(int))
    subset_sett2.scratch = <int *> sig_malloc((3*n+1) * sizeof(int))
    for i from 0 <= i < len(generators):
        for j from 0 <= j < n:
            gens[n*i + j] = generators[i][j]
    if SC_insert(supergroup, 0, gens, n_gens):
        raise MemoryError
    sig_free(gens)
    bitset_clear(&subset_sett1.bits)
    bitset_clear(&subset_sett2.bits)
    for i in set1:
        bitset_add(&subset_sett1.bits, i)
    for i in set2:
        bitset_add(&subset_sett2.bits, i)
    cdef bint isomorphic = sets_isom(supergroup, subset_sett1, subset_sett2, isom)
    SC_dealloc(supergroup)
    bitset_free(&subset_sett1.bits)
    bitset_free(&subset_sett2.bits)
    sig_free(subset_sett1.scratch)
    sig_free(subset_sett2.scratch)
    sig_free(subset_sett1)
    sig_free(subset_sett2)
    if isomorphic:
        output_py = [isom[i] for i from 0 <= i < n]
    else:
        output_py = False
    sig_free(isom)
    return output_py


cdef int sets_isom(StabilizerChain *supergroup, subset *set1, subset *set2, int *isom) except -1:
    r"""
    Underlying C function for testing two sets for isomorphism.
    """
    cdef int n = supergroup.degree
    cdef bint x
    cdef PartitionStack *part = PS_new(n, 1)
    if part is NULL:
        raise MemoryError
    x = double_coset(set1, set2, part, NULL, n,
        &all_set_children_are_equivalent, &refine_set, &compare_sets,
        supergroup, NULL, isom)
    PS_dealloc(part)
    return x

cdef bint all_set_children_are_equivalent(PartitionStack *PS, void *S) noexcept:
    return 0

cdef int refine_set(PartitionStack *PS, void *S, int *cells_to_refine_by, int ctrb_len) noexcept:
    """
    Given a set S, refine the partition stack PS so that each cell contains
    elements which are all either in the set or not in the set. If the depth is
    positive we do nothing since the cells will have already been split at an
    earlier level.
    """
    if PS.depth > 0:
        return 0
    cdef subset *subset1 = <subset *> S
    cdef int *scratch = subset1.scratch
    cdef int start, i, n = PS.degree
    start = 0
    while start < n:
        i = 0
        while True:
            scratch[i] = bitset_in(&subset1.bits, PS.entries[start+i])
            if PS.levels[start+i] <= PS.depth:
                break
            i += 1
        sort_by_function(PS, start, scratch)
        start += i + 1
    return 0

cdef inline int _bint_cmp(bint a, bint b) noexcept:
    return (<int> b) - (<int> a)

cdef int compare_sets(int *gamma_1, int *gamma_2, void *S1, void *S2, int degree) noexcept:
    r"""
    Compare two sets according to the lexicographic order.
    """
    cdef subset *subset1 = <subset *> S1
    cdef subset *subset2 = <subset *> S2
    cdef bitset_s set1 = subset1.bits
    cdef bitset_s set2 = subset2.bits
    cdef int i, j
    for i in range(degree):
        j = _bint_cmp(bitset_in(&set1, gamma_1[i]), bitset_in(&set2, gamma_2[i]))
        if j != 0:
            return j
    return 0

cdef void *allocate_subset(int n) noexcept:
    r"""
    Allocate a subset struct of degree n.
    """
    cdef subset *set1 = <subset *> sig_malloc(sizeof(subset))
    cdef int *scratch = <int *> sig_malloc((3*n+1) * sizeof(int))
    if set1 is NULL or scratch is NULL:
        sig_free(set1)
        sig_free(scratch)
        return NULL
    try:
        bitset_init(&set1.bits, n)
    except MemoryError:
        sig_free(set1)
        sig_free(scratch)
        return NULL
    set1.scratch = scratch
    return <void *> set1

cdef void free_subset(void *child) noexcept:
    r"""
    Deallocate a subset struct.
    """
    cdef subset *set1 = <subset *> child
    if set1 is not NULL:
        sig_free(set1.scratch)
        bitset_free(&set1.bits)
    sig_free(set1)

cdef void *allocate_sgd(int degree) noexcept:
    r"""
    Allocate the data part of an iterator which generates augmentations, i.e.,
    elements to add to the set.
    """
    cdef subset_generator_data *sgd = <subset_generator_data *> sig_malloc(sizeof(subset_generator_data))
    sgd.orbits = OP_new(degree)
    if sgd is NULL or sgd.orbits is NULL:
        deallocate_sgd(sgd)
        return NULL
    return <void *> sgd

cdef void deallocate_sgd(void *data) noexcept:
    r"""
    Deallocate the data part of the augmentation iterator.
    """
    cdef subset_generator_data *sgd = <subset_generator_data *> data
    if sgd is not NULL:
        OP_dealloc(sgd.orbits)
    sig_free(sgd)

cdef void *subset_generator_next(void *data, int *degree, bint *mem_err) noexcept:
    r"""
    Return the next element to consider adding to the set.
    """
    cdef subset_generator_data *sgd = <subset_generator_data *> data
    while True:
        sgd.cur_point += 1
        if sgd.cur_point == sgd.orbits.degree:
            break
        if OP_find(sgd.orbits, sgd.cur_point) == sgd.cur_point and \
          not bitset_in(&sgd.bits, sgd.cur_point):
            break
    if sgd.cur_point == sgd.orbits.degree or mem_err[0]:
        return NULL
    return <void *> &sgd.cur_point

cdef int generate_child_subsets(void *S, aut_gp_and_can_lab *group, iterator *child_iterator) noexcept:
    r"""
    Set up an iterator of augmentations, i.e., elements to add to the given set.
    """
    cdef subset *subset1 = <subset *> S
    cdef int i, j, n = group.group.degree
    cdef subset_generator_data *sgd = <subset_generator_data *> child_iterator.data
    OP_clear(sgd.orbits)
    for i in range(group.num_gens):
        for j in range(n):
            OP_join(sgd.orbits, j, group.generators[n*i + j])
    i = bitset_first(&subset1.bits)
    j = bitset_next(&subset1.bits, i+1)
    while j != -1:
        OP_join(sgd.orbits, i, j)
        j = bitset_next(&subset1.bits, j+1)
    sgd.cur_point = -1
    sgd.bits = subset1.bits
    return 0

cdef void *apply_subset_aug(void *parent, void *aug, void *child, int *degree, bint *mem_err) noexcept:
    r"""
    Add the element represented by ``aug`` to ``parent``, storing the result to
    ``child``.
    """
    cdef subset *set1 = <subset *> child
    cdef subset *par_set = <subset *> parent
    cdef bitset_s parbits = par_set.bits
    cdef int add_pt = (<int *> aug)[0], n = parbits.size
    bitset_copy(&set1.bits, &parbits)
    bitset_add(&set1.bits, add_pt)
    degree[0] = n
    return <void *> set1

cdef void free_subset_aug(void *aug) noexcept:
    return

cdef void *canonical_set_parent(void *child, void *parent, int *permutation, int *degree, bint *mem_err) noexcept:
    r"""
    Determine the canonical parent of the set ``child`` by applying
    ``permutation``, deleting the largest element in lexicographic order, and
    storing the result to ``parent``.
    """
    cdef subset *set1 = <subset *> child
    cdef int i, max_in_can_lab, max_loc, n = set1.bits.size
    cdef subset *par
    if parent is NULL:
        par = <subset *> allocate_subset(n)
        if par is NULL:
            mem_err[0] = 1
    else:
        par = <subset *> parent
    if mem_err[0]:
        return NULL
    i = bitset_first(&set1.bits)
    max_in_can_lab = permutation[i]
    max_loc = i
    while i != -1:
        if max_in_can_lab < permutation[i]:
            max_in_can_lab = permutation[i]
            max_loc = i
        i = bitset_next(&set1.bits, i+1)
    bitset_copy(&par.bits, &set1.bits)
    bitset_discard(&par.bits, max_loc)
    degree[0] = n
    return <void *> par

cdef iterator *allocate_subset_gen(int degree, int max_size) noexcept:
    r"""
    Allocate the generator of subsets.
    """
    cdef iterator *subset_gen = <iterator *> sig_malloc(sizeof(iterator))
    if subset_gen is not NULL:
        if allocate_subset_gen_2(degree, max_size, subset_gen):
            sig_free(subset_gen)
            subset_gen = NULL
    return subset_gen

cdef int allocate_subset_gen_2(int degree, int max_size, iterator *it) noexcept:
    r"""
    Given an already allocated iterator, allocate the generator of subsets.
    """
    cdef canonical_generator_data *cgd = allocate_cgd(max_size + 1, degree)
    if cgd is NULL:
        return 1
    cdef int i, j
    for i from 0 <= i < max_size + 1:
        cgd.object_stack[i] = allocate_subset(degree)
        cgd.parent_stack[i] = allocate_subset(degree)
        cgd.iterator_stack[i].data = allocate_sgd(degree)
        cgd.iterator_stack[i].next = &subset_generator_next
        if cgd.iterator_stack[i].data is NULL or \
           cgd.object_stack[i]        is NULL or \
           cgd.parent_stack[i]        is NULL:
            for j from 0 <= j <= i:
                deallocate_sgd(cgd.iterator_stack[i].data)
                free_subset(cgd.object_stack[i])
                free_subset(cgd.parent_stack[i])
            deallocate_cgd(cgd)
            return 1
    it.data = <void *> cgd
    it.next = canonical_generator_next
    return 0

cdef void free_subset_gen(iterator *subset_gen) noexcept:
    r"""
    Free the iterator of subsets.
    """
    if subset_gen is NULL:
        return
    cdef canonical_generator_data *cgd = <canonical_generator_data *> subset_gen.data
    deallocate_cgd(cgd)
    sig_free(subset_gen)

cdef iterator *setup_set_gen(iterator *subset_gen, int degree, int max_size) noexcept:
    r"""
    Initiate the iterator of subsets.
    """
    cdef subset *empty_set
    cdef iterator *subset_iterator = setup_canonical_generator(degree,
        &all_set_children_are_equivalent,
        &refine_set,
        &compare_sets,
        &generate_child_subsets,
        &apply_subset_aug,
        &free_subset,
        &deallocate_sgd,
        &free_subset_aug,
        &canonical_set_parent,
        max_size+1, 0, subset_gen)
    if subset_iterator is not NULL:
        empty_set = <subset *> (<canonical_generator_data *> subset_gen.data).object_stack[0]
        bitset_clear(&empty_set.bits)
    return subset_iterator


def sets_modulo_perm_group(list generators, int max_size,
                           bint indicate_mem_err=1):
    r"""
    Given generators of a permutation group, list subsets up to permutations in
    the group.

    INPUT:

    - ``generators`` -- list of generators in list form
    - ``max_size`` -- integer; maximum size of subsets to be generated
    - ``indicate_mem_err`` -- boolean; whether to raise an error.
      If we run out of memory, or simply append a :exc:`MemoryError`
      instance to the end of the output.

    EXAMPLES::

        sage: from sage.groups.perm_gps.partn_ref.refinement_sets import sets_modulo_perm_group
        sage: sets_modulo_perm_group([], 0)
        [[]]
        sage: sets_modulo_perm_group([], 1)
        [[0], []]
        sage: sets_modulo_perm_group([], 2)
        [[0, 1], [0], []]
        sage: sets_modulo_perm_group([], 3)
        [[0, 1, 2], [0, 1], [0], []]
        sage: sets_modulo_perm_group([], 4)
        [[0, 1, 2, 3], [0, 1, 2], [0, 1], [0], []]
        sage: len(sets_modulo_perm_group([], 99))
        100

    ::

        sage: sets_modulo_perm_group([[1,2,0]], 4)
        [[0, 1, 2], [0, 1], [0], []]
        sage: sets_modulo_perm_group([[1,2,0]], 3)
        [[0, 1, 2], [0, 1], [0], []]
        sage: sets_modulo_perm_group([[1,2,0]], 2)
        [[0, 1], [0], []]
        sage: sets_modulo_perm_group([[1,2,0]], 1)
        [[0], []]
        sage: sets_modulo_perm_group([[1,2,0]], 0)
        [[]]
        sage: sets_modulo_perm_group([[0,1,2]], 3)
        [[0, 1, 2], [0, 1], [0, 2], [0], [1, 2], [1], [2], []]
        sage: sets_modulo_perm_group([[1,0,2]], 3)
        [[0, 1, 2], [0, 1], [0, 2], [0], [2], []]
        sage: sets_modulo_perm_group([[1,0,2],[1,2,0]], 3)
        [[0, 1, 2], [0, 1], [1], []]

    ::

        sage: sets_modulo_perm_group([[1,2,3,0]], 4)
        [[0, 1, 2, 3], [0, 1, 2], [0, 1], [0, 2], [0], []]
        sage: sets_modulo_perm_group([[1,2,3,0]], 5)
        [[0, 1, 2, 3], [0, 1, 2], [0, 1], [0, 2], [0], []]
        sage: sets_modulo_perm_group([[1,2,3,0]], 3)
        [[0, 1, 2], [0, 1], [0, 2], [0], []]
        sage: sets_modulo_perm_group([[1,2,3,0]], 2)
        [[0, 1], [0, 2], [0], []]
        sage: sets_modulo_perm_group([[1,2,3,0]], 1)
        [[0], []]
        sage: sets_modulo_perm_group([[1,2,3,0]], 0)
        [[]]
        sage: sets_modulo_perm_group([[0,1,3,2],[1,0,2,3]], 4)
        [[0, 1, 2, 3], [0, 1, 2], [0, 1], [0, 2, 3], [0, 2], [0], [2, 3], [2], []]
        sage: sets_modulo_perm_group([[1,0,2,3],[1,2,0,3]], 4)
        [[0, 1, 2, 3], [0, 1, 2], [0, 1, 3], [0, 1], [1, 3], [1], [3], []]
        sage: sets_modulo_perm_group([[1,2,0,3],[0,2,3,1]], 4)
        [[0, 1, 2, 3], [0, 1, 2], [0, 1], [1], []]
        sage: sets_modulo_perm_group([[1,0,2,3],[1,2,3,0]], 4)
        [[0, 1, 2, 3], [0, 1, 2], [1, 2], [2], []]
        sage: L = list(powerset(range(4)))
        sage: L.sort()
        sage: L == sorted(sets_modulo_perm_group([[0,1,2,3]], 4))
        True

    ::

        sage: sets_modulo_perm_group([[1,2,3,4,0]], 5)
        [[0, 1, 2, 3, 4], [0, 1, 2, 3], [0, 1, 2], [0, 1], [0, 2, 3], [0, 2], [0], []]
        sage: sets_modulo_perm_group([[1,0,2,3,4],[0,1,3,4,2]], 5)
        [[0, 1, 2, 3, 4], [0, 1, 2, 3], [0, 1, 2], [0, 1], [0, 2, 3, 4], [0, 2, 3], [0, 2], [0], [2, 3, 4], [2, 3], [2], []]
        sage: L = list(powerset(range(5)))
        sage: L.sort()
        sage: L == sorted(sets_modulo_perm_group([[0,1,2,3,4]], 5))
        True
        sage: sets_modulo_perm_group([[1,0,2,3,4],[1,2,3,4,0]], 5)
        [[0, 1, 2, 3, 4], [0, 1, 2, 3], [1, 2, 3], [2, 3], [3], []]
        sage: sets_modulo_perm_group([[1,2,0,3,4],[0,2,3,1,4],[0,1,3,4,2]], 5)
        [[0, 1, 2, 3, 4], [0, 1, 2, 3], [1, 2, 3], [1, 2], [2], []]

    ::

        sage: X = sets_modulo_perm_group([[1,2,3,4,5,0]], 6)
        sage: [a for a in X if len(a) == 0]
        [[]]
        sage: [a for a in X if len(a) == 1]
        [[0]]
        sage: [a for a in X if len(a) == 2]
        [[0, 1], [0, 2], [0, 3]]
        sage: [a for a in X if len(a) == 3]
        [[0, 1, 2], [0, 1, 3], [0, 2, 3], [0, 2, 4]]
        sage: [a for a in X if len(a) == 4]
        [[0, 1, 2, 3], [0, 1, 3, 4], [0, 2, 3, 4]]
        sage: [a for a in X if len(a) == 5]
        [[0, 1, 2, 3, 4]]
        sage: [a for a in X if len(a) == 6]
        [[0, 1, 2, 3, 4, 5]]

    ::

        sage: X = sets_modulo_perm_group([[0,2,1,4,3,5,8,7,6],[8,7,6,3,5,4,2,1,0]], 9)
        sage: len(X)
        74
    """
    cdef list out_list = []
    cdef int i
    if max_size == 0:
        return [[]]
    if len(generators) == 0:
        ll = []
        for i in range(max_size,-1,-1):
            ll.append(list(range(i)))
        return ll
    cdef int n = len(generators[0]), n_gens = len(generators)
    cdef iterator *subset_iterator
    cdef subset *thing

    cdef StabilizerChain *group = SC_new(n)
    cdef int *gens = <int *> sig_malloc(n*n_gens * sizeof(int))
    if group is NULL or gens is NULL:
        SC_dealloc(group)
        sig_free(gens)
        raise MemoryError
    for i from 0 <= i < len(generators):
        for j from 0 <= j < n:
            gens[n*i + j] = generators[i][j]
    if SC_insert(group, 0, gens, n_gens):
        SC_dealloc(group)
        sig_free(gens)
        raise MemoryError
    sig_free(gens)

    cdef iterator *subset_gen = allocate_subset_gen(n, max_size)
    if subset_gen is NULL:
        SC_dealloc(group)
        raise MemoryError
    subset_iterator = setup_set_gen(subset_gen, n, max_size)
    cdef bint mem_err = 0
    if subset_iterator is NULL:
        SC_dealloc(group)
        free_subset_gen(subset_gen)
        mem_err = 1
    else:
        start_canonical_generator(group, NULL, n, subset_gen)
    while not mem_err:
        thing = <subset *> subset_iterator.next(subset_iterator.data, NULL, &mem_err)
        if thing is NULL:
            break
        out_list.append( bitset_list(&thing.bits) )
    free_subset_gen(subset_gen)
    SC_dealloc(group)
    if mem_err:
        if indicate_mem_err:
            raise MemoryError
        else:
            out_list.append(MemoryError())
    return out_list
